open Ast_types
open Java_context_types


let rec expr e =
  match e.expr_node with 
  | Lit _ -> []
  | Bin(e1, _, e2) -> (expr e1)@(expr e2)
  | Un(_, e) -> expr e
  | PrePostIncrDecr(_, e) -> expr e
  | Typed_var _ -> []
  | Static_class _ -> []
  | Static_interface _ -> []
  | Assign_local_var(_, _, e) -> expr e
  | Assign_field(_, _, e) -> expr e
  | Typed_assign_array(_, e1, e2, _, e3) -> (expr e1)@(expr e2)@(expr e3)
  | If_expr(e1, e2, e3) -> (expr e1)@(expr e2)@(expr e3)
  | This -> []
  | Field_access _ -> []
  | Typed_method_call(e, me, el) -> 
      let el = List.fold_left (fun acc e -> expr e) [] el in
      let me = if me.method_entry_routine.routine_entry_produce_body then [Method_entry [me]] else [] in
      (expr e)@me@el
  | Typed_class_instance_creation(ce, el) ->
      let el = List.fold_left (fun acc e -> expr e) [] el in
      let ce = if ce.constructor_entry_routine.routine_entry_produce_body then [Constructor_entry ce] else [] in
      ce@el
  | Array_creation _ -> []
  | Typed_array_access(_, e1, e2) -> (expr e1)@(expr e2)
  | Typed_cast(_, _, e) -> expr e
  | Typed_instanceof(e, _) -> expr e
  | _ -> []
;;


let rec variable_initializer vi =
  match vi with
  | Simple_initializer e -> expr e
  | Array_initializer vil -> 
      List.fold_left (fun acc vi -> variable_initializer vi) [] vil
;;

let rec statement s = 
  match s.statement_node with  
    | Skip -> []
    | Typed_expr(_, e) -> expr e
    | Typed_var_decl(_, None, s) -> statement s
    | Typed_var_decl(_, Some vi, s) -> (variable_initializer vi)@(statement s)
    | Throw e -> expr e								   
    | Typed_return(_, None) -> []
    | Typed_return(_, Some e) -> expr e
    | Break _ -> []
    | Continue _ -> []
    | Label(_, s) -> statement s
    | If_statement (e, s1, s2) -> (expr e)@(statement s1)@(statement s2)
    | While(_, e, s) -> (expr e)@(statement s)
    | Do(_, s, e) -> (expr e)@(statement s)
    | For(_, sl1, e, sl2, s) ->
	let al1 = List.fold_left (fun acc s -> statement s) [] sl1 in
	let al2 = List.fold_left (fun acc s -> statement s) [] sl2 in
	al1@(expr e)@al2@(statement s)
    | For_typed_decl(_, None, _, e, sl, s) ->
	let al = List.fold_left (fun acc s -> statement s) [] sl in
	(expr e)@al@(statement s)
    | For_typed_decl(_, Some vi, _, e, sl, s) ->
	let al = List.fold_left (fun acc s -> statement s) [] sl in
	(variable_initializer vi)@(expr e)@al@(statement s)
    | Typed_try(b, catches, None) ->
	let catches = List.fold_left (fun acc (_, _, b) -> block b) [] catches in
	(block b)@catches
    | Typed_try(b, catches, Some finally) -> 
	let catches = List.fold_left (fun acc (_, _, b) -> block b) [] catches in
	(block b)@catches@(block finally)
    | Switch(e, cases) -> 
 	let cases = List.fold_left (fun acc (_, b) -> block b) [] cases in
	(expr e)@cases
    | Block b -> block b
    | Synchronized(e, b) -> (expr e)@(block b)
    | Typed_annotated(_, s) -> statement s
    | _ -> []


and block b =
  List.fold_left (fun acc s -> statement s) [] b
;;


let make_graph fd = 
  match fd with    
  | Typed_method(me, _, Some b) -> 
      me.method_entry_graph <- (block b);
  | Typed_constructor(ce, ms, eci, b) ->
      ce.constructor_entry_graph <- (block b);
  | _ -> ()
;;
      

let compilation_unit cu = 
  List.iter 
    (fun td -> 
      match td with
      | Typed_class(ce, cd) -> 
	  List.iter (fun fd ->  make_graph fd) (List.map snd cd.class_fields)
      | _ -> ())
    (List.map snd cu.cu_type_decls) 
;;


(* Pretty Print *)

let rec print_graph env = 
  List.iter
    (fun (_, jee) -> 
      match jee with
      | Package_entry pe -> print_graph pe.package_entry_contents
      | Class_entry ce -> print_graph ce.class_entry_contents
      | Method_entry [me] -> 
	Format.eprintf "%s :"  me.method_entry_trans_name;
	  List.iter 
	    (fun jee -> 
	      match jee with
	      | Method_entry [me] -> Format.eprintf " %s," me.method_entry_trans_name
	      | Constructor_entry ce -> Format.eprintf " %s," ce.constructor_entry_trans_name
	      | _ -> assert false) 
	    me.method_entry_graph;
	  Format.eprintf "@\n"
      | Constructor_entry ce -> 
	  Format.eprintf "%s :"  ce.constructor_entry_trans_name;
	  List.iter 
	    (fun jee -> 
	      match jee with
	      | Method_entry [me] -> Format.eprintf " %s," me.method_entry_trans_name
	      | Constructor_entry ce -> Format.eprintf " %s," ce.constructor_entry_trans_name
	      | _ -> assert false) 
	    ce.constructor_entry_graph;
	  Format.eprintf "@\n"
      | _ -> ()) 
    env
    
  
(*
open Info
  



module G = struct 
  type t = (string*Cast.tfile) list
  module V = struct
    type t = fun_info
    let compare f1 f2 = Pervasives.compare f1.fun_tag f2.fun_tag
    let hash f = f.fun_tag 
    let equal f1 f2 = f1.fun_tag == f2.fun_tag
  end
  let iter_vertex iter files =
    (*let iter_fun  e = 
      match e with    
	| Tfundef (s, t, f, _)  -> iter f
	| _ -> () in*)
      Cenv.iter_sym 
	(fun n x -> match x with
	   | Var_info _ -> () 
	   | Fun_info f -> iter f) 
    (*List.iter (fun (_,file) ->List.iter (fun d -> iter_fun d.node) file) files*)
  let iter_succ iter _ f =
    List.iter iter f.graph 
  end

module SCC = Components.Make(G)

let find_comp tfiles =  
  let (scc,numcomp) = SCC.scc tfiles in
  let tab_comp = Array.make numcomp [] in
  G.iter_vertex 
    (fun f ->
       let i = scc f in 
       Array.set tab_comp i (f::(Array.get tab_comp i))) tfiles;
  Coptions.lprintf "Call graph by components: @.";
  Array.iteri (fun i l -> Coptions.lprintf "%d: " i;
		 List.iter (fun f -> Coptions.lprintf "%s " f.fun_name) l;
		 Coptions.lprintf "@.")
    tab_comp;
  tab_comp
*)
