(***************************************************************************

Java context 

VerifiCard Project - Dmons research team - LRI - Universit Paris XI

$Id: java_context.ml,v 1.34 2007/03/11 14:28:40 nrousset Exp $

***************************************************************************)



open Java_context_types;;

(*

  miscellaneous

*)



let get_ive_class_or_interface_name ive =
  match ive.instance_variable_entry_class_or_interface with
    | Class_entry ce -> ce.class_entry_name
    | Interface_entry ie -> ie.interface_entry_name 
    | _ -> assert false (* should never occur *)
;;


let rec equals_java_type t1 t2 =
  match t1, t2 with
    | (Integer_type, Integer_type)
    | (Float_type, Float_type)
    | (Boolean_type, Boolean_type)
    | (Null_type, Null_type)
    | (Type_type, Type_type) -> true
    | (Class_type ce1, Class_type ce2) -> ce1.class_entry_name = ce2.class_entry_name (* WARNING: false until name resolution *)
    | (Interface_type ie1, Interface_type ie2) -> ie1.interface_entry_name = ie2.interface_entry_name (* IDEM *)
    | (Array_type t1, Array_type t2) -> equals_java_type t1 t2
    | _ -> false
;;

let is_ref_type t = 
  match t with
  | Byte_type | Short_type | Integer_type 
  | Float_type | Boolean_type | Real_type -> false
  | Null_type -> assert false
  | Type_type -> assert false
  | Class_type _ | Interface_type _ | Array_type _ -> true
  | Type_var _
  | Logic_type(_, _)
  | Prop_type -> assert false

let type_entry_name t = 
  match t with
    | Class_type ce  -> ce.class_entry_name
    | Interface_type ie -> ie.interface_entry_name
    | _ -> assert false
;;

let default_effects =
  {
    reads_alloc = false;
    writes_alloc = false;
    writes_everything = false;
    reads_alloc_table = false;
    writes_alloc_table = false;
    reads = Inttagset.empty;
    writes = Inttagset.empty;
    reads_int_array = false;
    reads_int_assigned_array = false;
    writes_int_array = false;
    reads_float_array = false;
    writes_float_array = false;
    reads_bool_array = false;
    writes_bool_array = false;
    reads_obj_array = false;
    writes_obj_array = false;
    throws = false (*I Inttagset.empty i*);
    breaks = false;
    continue = false
  }
    
let default_routine =
  {
   routine_entry_parameters = [];
   routine_entry_parameters_as_local_var = [];
   routine_entry_produce_body = false;
   routine_entry_local_vars = [];
   routine_entry_int_array_writes_nothing = false;
   routine_entry_float_array_writes_nothing = false;
   routine_entry_bool_array_writes_nothing = false;
   routine_entry_obj_array_writes_nothing = false;
   routine_entry_fields_writes_nothing = Inttagset.empty;
   routine_entry_effects = default_effects;
   routine_entry_use_transactions = false;
 }
    
(*

  printing

*)

open Format;;
open Format_misc;;


let rec fprintf_package_entry form pe =
  fprintf form "@[<hv 2>{ @[<hv 1>name =@ %s;@]@ @[<hv 1>dir =@ %a;@]@ @[<hv 1>contents =@ %a;@] }@]"
    pe.package_entry_name 
    (fprintf_list (fun form s -> fprintf form "%s@." s)) 
    pe.package_entry_directories
    fprintf_java_context pe.package_entry_contents

and fprintf_class_entry form ce =
  fprintf form "@[<hv 2>{ name =@ %s;@ <...>;@ contents =@ %a;@ }@]"
    ce.class_entry_name 
    fprintf_java_context ce.class_entry_contents

and fprintf_interface_entry form ce =
  fprintf form "@[<hv 2>{ name =@ %s;@ <...>;@ contents =@ %a;@ }@]"
    ce.interface_entry_name 
    fprintf_java_context ce.interface_entry_contents

and fprintf_local_variable_entry form ve =
  fprintf form "@[<hv 2>{ name =@ %s;@ type =@ %a;@ }@]"
    ve.local_variable_entry_name
    fprintf_java_type ve.local_variable_entry_type

and fprintf_logic_type_entry form lte =
  fprintf form "@[<hv 2>{ name =@ %s;@ }@]"
    lte.logic_type_entry_name

and fprintf_instance_variable_entry form ve =
  fprintf form "@[<hv 2>{ name =@ %s; is_static =@ %b;@ type =@ %a;@ }@]"
    ve.instance_variable_entry_name
    ve.instance_variable_entry_is_static
    fprintf_java_type ve.instance_variable_entry_type

and fprintf_constant_entry form ve =
  fprintf form "@[<hv 2>{ name =@ %s;@ type =@ %a;@ }@]"
    ve.constant_entry_name
    fprintf_java_type ve.constant_entry_type

and fprintf_method_entry form me =
  fprintf form "@[<hv 2>{ name =@ %s;@ is_static =@ %b;@ is_pure =@ %b;@ return_type =@ %a;@ }@]"
    me.method_entry_name
    me.method_entry_is_static
    me.method_entry_is_pure
    fprintf_java_type me.method_entry_return_type

and fprintf_constructor_entry form ce =
  fprintf form "@[<hv 2>{ name =@ %s;@ <...>;@ }@]"
    ce.constructor_entry_trans_name

and fprintf_java_type form t =
  match t with
    | Byte_type -> fprintf form "byte"
    | Short_type -> fprintf form "short" 
    | Integer_type -> fprintf form "int"
    | Float_type -> fprintf form "float"
    | Real_type -> fprintf form "real"
    | Boolean_type -> fprintf form "bool"
    | Null_type -> fprintf form "unit"
    | Type_var s -> fprintf form "%s" s
    | Logic_type(id, idl) -> 
	fprintf form "logic(%s, %a)"
	  id
	  (fprintf_list fprintf_java_type) idl
    | Prop_type -> fprintf form "prop"
    | Class_type ce ->
	fprintf form "class(%s)" ce.class_entry_name
    | Interface_type ie ->
	fprintf form "interface(%s)" ie.interface_entry_name
    | Array_type t ->
	fprintf form "%a[]" fprintf_java_type t
    | Type_type -> assert false

and fprintf_java_env_entry form e =
  match e with
  | Package_entry pe -> 
      fprintf form "@[<hv 1>Package_entry@,(%a)@]" fprintf_package_entry pe
  | Class_entry te ->
      fprintf form "@[<hv 1>Class_entry@,(%a)@]" fprintf_class_entry te       
  | Interface_entry te ->
      fprintf form "@[<hv 1>Interface_entry@,(%a)@]" fprintf_interface_entry te       
  | Instance_variable_entry ive ->
      fprintf form "@[<hv 1>Instance_variable_entry@,(%a)@]" 
	fprintf_instance_variable_entry ive
  | Constant_entry ce ->
      fprintf form "@[<hv 1>Constant_entry@,(%a)@]" fprintf_constant_entry ce
  | Method_entry me ->
      fprintf form "@[<hv 1>Method_entry@,(%a)@]" 
	(fprintf_list fprintf_method_entry) me
  | Constructor_entry ce ->
      fprintf form "@[<hv 1>Constructor_entry@,(%a)@]" 
	fprintf_constructor_entry ce
  | Local_variable_entry lve ->
      fprintf form "@[<hv 1>Local_variable_entry@,(%a)@]" 
	fprintf_local_variable_entry lve
  | Logic_type_entry lte ->
      fprintf form "@[<hv 1>Logic_type@,(%a)@]"
        fprintf_logic_type_entry lte
  | Logic_entry le ->
      fprintf form "@[<hv 1>Logice@,(%s,...)@]" le.logic_entry_name

and fprintf_java_context form l =
  fprintf_list 
    (fun form (id,e) -> 
       fprintf form "@[<hv 1>%s =>@ %a;@]" id fprintf_java_env_entry e)
    form l;
;;

let printf_java_context = fprintf_java_context Format.std_formatter;;


let fprintf_iveset form iveset =
  fprintf form "@[{";
  Inttagset.iter
    (fun ive -> fprintf form "%s@ " ive.instance_variable_entry_name)
    iveset;
  fprintf form "}@]"
;;
  


let fprintf_effects form e =
  fprintf form "@[<hv 2>{ ";
  fprintf form "reads_alloc = %b ;@ " e.reads_alloc;
  fprintf form "writes_alloc = %b ;@ " e.writes_alloc;
  fprintf form "reads_alloc_table = %b ;@ " e.reads_alloc_table;
  fprintf form "writes_alloc_table = %b ;@ " e.writes_alloc_table;
  fprintf form "reads = %a;@ " fprintf_iveset e.reads;
  fprintf form "writes = %a;@ " fprintf_iveset e.writes;
  fprintf form "reads_int_array = %b ;@ " e.reads_int_array;
  fprintf form "writes_int_array = %b ;@ " e.writes_int_array;
  fprintf form "reads_float_array = %b ;@ " e.reads_float_array;
  fprintf form "writes_float_array = %b ;@ " e.writes_float_array;
  fprintf form "reads_bool_array = %b ;@ " e.reads_bool_array;
  fprintf form "writes_bool_array = %b ;@ " e.writes_bool_array;
  fprintf form "reads_obj_array = %b ;@ " e.reads_obj_array;
  fprintf form "writes_obj_array = %b ;@ " e.writes_obj_array;
  fprintf form "throws = %b ;@ " e.throws (*I Inttagset.empty i*);
  fprintf form "breaks = %b ;@ " e.breaks;
  fprintf form "}@]";


