(***************************************************************************

Lexer for Jml assertions

Krakatoa Project - Dmons research team - LRI - Universit Paris XI

$Id: jml_lexer.mll,v 1.22 2007/03/12 15:05:03 marche Exp $

***************************************************************************)


{

  open Jml_parser

  exception Lexical_error of string

  let string_buffer = Buffer.create 17

  let kw_table = 
    let table = Hashtbl.create 17 in
    let _ = 
      List.iter
	(fun (s,t) -> Hashtbl.add table s t)
	[ "also", ALSO;
	  "and", AND;
	  "assert", ASSERT ;
	  "assignable", ASSIGNABLE;
	  "assigns", ASSIGNABLE;
	  "axiom", AXIOM;
	  "behavior", BEHAVIOR;
	  "boolean", BOOLEAN;
	  "byte", BYTE;
          "call_sequence", CALL_SEQUENCE;
          "cardtear", CARDTEAR;
	  "decreases", DECREASES;
	  "diverges", DIVERGES;
	  "double", DOUBLE;
	  "ensures", ENSURES;
	  "ensures_redundantly", ENSURES;
	  "exceptional_behavior", EXCEPTIONAL_BEHAVIOR;
	  "false", FALSE(Location.dummy);
	  "final", FINAL;
	  "float", FLOAT;
	  "forall", FORALL;
	  "ghost", GHOST;
	  "if", IF;
	  "instanceof", INSTANCEOF;
	  "int", INT;
	  "invariant", INVARIANT;
	  "logic", LOGIC;
	  "long", LONG;
	  "loop_invariant", LOOP_INVARIANT;
	  "measured_by", MEASURED_BY;
	  "model", MODEL;
	  "modifiable", ASSIGNABLE;
	  "modifies", ASSIGNABLE;
	  "new", NEW(Location.dummy);
	  "normal_behavior", NORMAL_BEHAVIOR;
	  "null", NULL(Location.dummy);
	  "old", OLD;
	  "predicate", PREDICATE;
	  "private", PRIVATE;	
	  "protected", PROTECTED;
	  "public", PUBLIC;
	  "pure", PURE;
	  "reads", READS;
	  "real", REAL_TYPE;
	  "representation", REPRESENTATION;
	  "represents", REPRESENTS;
	  "requires", REQUIRES;
          "requires_redundantly", REQUIRES; 
	  "set", SET;
	  "short", SHORT;
	  "signals", SIGNALS;
	  "spec_public", SPEC_PUBLIC;
	  "static", STATIC;
	  "super", SUPER(Location.dummy);
	  "true", TRUE(Location.dummy);
	  "type", TYPE;
	  "when", WHEN;
	]
    in table


  let special_kw_table = 
    let table = Hashtbl.create 17 in
    let _ = 
      List.iter
	(fun (s,t) -> Hashtbl.add table s t)
	[ "everything", BSEVERYTHING ;
	  "exists", BSEXISTS ;
	  "fresh", BSFRESH(Location.dummy) ;
	  "forall", BSFORALL ;
	  "nothing", BSNOTHING;
	  "fields_of", BSFIELDSOF;
          "not_conditionally_updated", BSNOTCONDITIONALLYUPDATED;
	  "old", BSOLD(Location.dummy) ;
	  "result", BSRESULT(Location.dummy) ;
	  "type", BSTYPE(Location.dummy) ;
	  "typeof", BSTYPEOF(Location.dummy) ;
	  "fpi", BSFPI(Location.dummy) ;
	]
    in table

  let find_loc lexbuf =
    Location.find 
      !Jml_parser_base.cur_base 
      (Lexing.lexeme_start lexbuf) 
      (Lexing.lexeme_end lexbuf) 
      
  let id_or_kw lexbuf s =
    let l = find_loc lexbuf in
    try
      let k = Hashtbl.find kw_table s in
      (*i
	prerr_string "Keyword ";
	prerr_endline s;
	i*)
      begin
	match k with
	  | FALSE _ -> FALSE(l)
	  | NEW _ -> NEW(l)
	  | NULL _ -> NULL(l)
	  | SUPER _ -> SUPER(l)
	  | THIS _ -> THIS(l)
	  | TRUE _ -> TRUE(l)
	  | _ -> k
      end
    with
	Not_found ->
	  (*i
	    prerr_string "Ident ";
	    prerr_endline s;
	    i*)
	  ID(l,s)

  let special_id lexbuf s =
    let l = find_loc lexbuf in
   try
      let k = Hashtbl.find special_kw_table s in
      (*i
	prerr_string "Special Keyword ";
	prerr_endline s;
	i*)
      begin
	match k with
	  | BSRESULT _ -> BSRESULT(l)
	  | BSOLD _ -> BSOLD(l)
	  | BSFPI _ -> BSFPI(l)
	  | BSTYPE _ -> BSTYPE(l)
	  | BSTYPEOF _ -> BSTYPEOF(l)
	  | _ -> k
      end
    with
	Not_found ->
	  (*i
	    prerr_string "Ident ";
	    prerr_endline s;
	    i*)
	  raise (Lexical_error ("unknown special JML keyword \\"^s)) 


   let remove_first_char s = String.sub s 1 (pred (String.length s))
   let remove_last_char s = String.sub s 0 (pred (String.length s))
   let remove_last_two_chars s = String.sub s 0 (pred (pred (String.length s)))

   exception Dotdot of Location.t * int

}

rule token = parse
  | [' ' '\t'] +
      { token lexbuf }
  | '\n' | '\r' | "\r\n"
      { Jml_parser_base.next_line lexbuf; token lexbuf }
  | "/*"
      { comment lexbuf }
  | "//" [^'\n']* '\n'
      { Jml_parser_base.next_line lexbuf; token lexbuf }
  | "\\" ['A'-'Z''a'-'z''_']['A'-'Z''a'-'z''0'-'9''_']* 
      { special_id lexbuf (remove_first_char (Lexing.lexeme lexbuf)) }
  | ['A'-'Z''a'-'z''_']['A'-'Z''a'-'z''0'-'9''_']* 
      { id_or_kw lexbuf (Lexing.lexeme lexbuf) }
  | ';'
      { SEMICOLON }
  | ','
      { COMMA }
  | '.'
      { DOT }
  | ".."
      { DOTDOT }
  | '+'                 
      { PLUS(find_loc lexbuf) }
  | '-'       
      { MINUS(find_loc lexbuf) }
  | '''
      { PRIME }
  | '*'
      { STAR }
  | '/'
      { SLASH }
  | '%'
      { PERCENT }
  | "&"
      { AMPERSAND }
  | "|"
      { VERTICALBAR }
  | "&&"
      { AMPERSANDAMPERSAND }
  | "||"
      { VERTICALBARVERTICALBAR }
  | "<-"
      { LEFTARROW }
  | "==>"
      { EQEQGT } 
  | "<==" 
      { LTEQEQ  }
  | "<==>"
      { LTEQEQGT }
  | "<=!=>"
      { LTEQBANGEQGT }
  | "!"
      { BANG(find_loc lexbuf) }
  | "~"
      { TILDA(find_loc lexbuf) }
  | "^"
      { CARET }
  | "?"
      { QUESTIONMARK }
  | ":"
      { COLON }
  | "<<" | ">>" | ">>>"
      { SHIFT(Lexing.lexeme lexbuf) }
  | ">" | "<" | "<=" | ">=" | "<:"
      { COMP(Lexing.lexeme lexbuf) }
  | "==" | "!="
      { EQOP(Lexing.lexeme lexbuf) }
  | '='
      { ASSIGN } 
  | "*=" | "/=" | "%=" | "+=" | "-=" 
  | "<<=" | ">>=" | ">>>=" | "&=" | "^=" | "|="
      { ASSIGNOP(Lexing.lexeme lexbuf) }

      (* decimal constants *)

  | '0' | ['1'-'9']['0'-'9']*
      { INTEGER(find_loc lexbuf,int_of_string (Lexing.lexeme lexbuf)) }
  | ('0' | ['1'-'9']['0'-'9']*) ['l''L']
      { INTEGER 
	  (find_loc lexbuf,int_of_string (remove_last_char (Lexing.lexeme lexbuf))) }

      (* octal constants *)

  | '0'['0'-'7']+          
      { INTEGER (find_loc lexbuf,int_of_string ("0o" ^ (Lexing.lexeme lexbuf))) }
  | '0'['0'-'7']+ ['l''L']         
      { INTEGER 
	  (find_loc lexbuf,int_of_string 
	     ("0o" ^ (remove_last_char (Lexing.lexeme lexbuf)))) }

      (* hexadecimal constants *)

  | '0'['x''X']['0'-'9' 'A'-'F' 'a'-'f']+          
    { INTEGER (find_loc lexbuf,int_of_string (Lexing.lexeme lexbuf)) }
  | '0'['x''X']['0'-'9' 'A'-'F' 'a'-'f']+['l''L'] 
    { INTEGER (find_loc lexbuf,int_of_string (remove_last_char (Lexing.lexeme lexbuf))) }

      (* floating-point constants *)

  | (['0'-'9']+ as _int) '.' (['0'-'9']* as _frac) 
      (['e''E'](['-''+']?['0'-'9']+ as _exp))? ['f''F''d''D'] ?
      { REAL (find_loc lexbuf,Lexing.lexeme lexbuf) }

  | '.' (['0'-'9']+ as _frac) (['e''E'](['-''+']?['0'-'9']+ as _exp))? ['f''F''d''D'] ?
      { REAL (find_loc lexbuf,Lexing.lexeme lexbuf) }

  | (['0'-'9']+ as _int) ['e''E'] (['-''+']?['0'-'9']+ as _exp) ['f''F''d''D'] ?
      { REAL (find_loc lexbuf,Lexing.lexeme lexbuf) }

      (* a workaround for the [5..10] problem... *)

  | ['0'-'9']+ ".."
      { raise 
	  (Dotdot
	     (find_loc lexbuf,int_of_string 
		(remove_last_two_chars (Lexing.lexeme lexbuf)))) }

      (* character constants *)

  | "'" _ "'"
      { CHARACTER (find_loc lexbuf,Lexing.lexeme_char lexbuf 1) }

  | '('
      { LEFTPAR(find_loc lexbuf) }
  | ')'
      { RIGHTPAR }
  | '{'
      { LEFTBRACE }
  | '}'
      { RIGHTBRACE }
  | "{|"
      { LEFTBRACEBAR }
  | "|}"
      { BARRIGHTBRACE }
  | '['
      { LEFTBRACKET }
  | ']'
      { RIGHTBRACKET }
  | '"'
      { let start = Lexing.lexeme_start lexbuf in
	Buffer.clear string_buffer;
	let _ = string lexbuf in
	let eend = Lexing.lexeme_end lexbuf in
	STRING(Location.find !Jml_parser_base.cur_base start eend,
	       Buffer.contents string_buffer) }
  | "@"
      { token lexbuf }
  | _ 
      { raise (Lexical_error (Lexing.lexeme lexbuf)) }
  | eof
      { EOF }

and string = parse
  | '"'
      { find_loc lexbuf }
  | '\\' ['n'] 
      { Buffer.add_string string_buffer (Lexing.lexeme lexbuf);
	string lexbuf }
  | [^'\\''"']+ 
      { Buffer.add_string string_buffer (Lexing.lexeme lexbuf);
	string lexbuf }
  | '\n' | eof
      { raise (Lexical_error "string not terminated") }

and comment = parse
  | "*/"                
      { token lexbuf }
  | '\n'
      { Jml_parser_base.next_line lexbuf; comment lexbuf }
  | [^'*''\n']+             
      { comment lexbuf }
  | _                   
      { comment lexbuf }
  | eof                 
      { raise (Lexical_error "comment not terminated") }



{

let dotdot_mem = ref false
 
let next_token lexbuf =
  if !dotdot_mem then
    begin
      dotdot_mem := false;
      DOTDOT
    end
  else
    begin
      try
	token lexbuf
      with
	Dotdot(l,n) ->
	  dotdot_mem := true;
	  INTEGER(l,n)
    end
	      
} 
