

(***************************************************

  Typing functions

*************************************************************)

open Java_context_types;;
open Java_context;;
open Ast_types;;
open Ast;;


let current_routine = ref default_routine;;

let new_tag =
  let c = ref 0 in
  fun () -> incr c; !c;;

let sprintf = Format.sprintf;;

exception Typing_error of string;;

let report_typing_error loc msg =
  if !Config.stop_on_error
  then
    raise 
      (Typing_error ((Location.to_string loc)^msg))
  else
    Format.printf "Typing error: %s%s@."
     (Location.to_string loc) msg
;;

exception Ill_typed of string;;

let ill_typed loc msg = raise (Ill_typed ((Location.to_string loc)^": "^msg));;

(***************************

Phase 0

reading contents of input_path variable

***************************)

let additional_files_to_parse = ref [];;

(*
let rec add_to_list x l =
  match l with
    | [] -> [x]
    | y::r -> if x=y then l else y::(add_to_list x r)
;;
*)

let already_parsed = ref [];;

let rec do_read_dir d dirh env =
  try
    let name = Unix.readdir dirh in
    if name <> "." && name <> ".."  && name <> "CVS" then
      let fullname = Filename.concat d name in
       let s = Unix.lstat fullname in
      if s.Unix.st_kind = Unix.S_DIR then
	let p = List.filter (fun (n, _) -> n = name) env in 
	match p with
	| [] ->
	let pack =
	  {
	   package_entry_name = name;
	   package_entry_directories = [fullname];
	   package_entry_contents_read = false;
	   package_entry_contents = [];
	 } 
	in 
	Format.fprintf Config.log "Adding package %s in environment@."
	  fullname;
	do_read_dir d dirh 
	  ((name, Package_entry pack)::env)
	| [(_, Package_entry pe)] ->
	    pe.package_entry_directories <-
	      fullname::pe.package_entry_directories;
	    do_read_dir d dirh env
	| _ -> assert false (* should never occur *)
      else if s.Unix.st_kind = Unix.S_REG then
	let base,ext = Config.split_extension name in
	match ext with
	| "java" | "java-refined" | "jml" | "jml-refined" ->
	    if !Config.verbose >= 1 then
	      Format.printf "Adding class or interface %s in environment@."
	      fullname;
            (*i*)
            if Filename.basename d <> Filename.basename !Config.source_package || not (List.mem base !already_parsed)
            then 
	    additional_files_to_parse :=
	    (* add_to_list fullname !additional_files_to_parse; *)
	      fullname:: !additional_files_to_parse;                
	    (*i*)
            if Filename.basename d = Filename.basename !Config.source_package && not (List.mem base !already_parsed)
            then already_parsed := base::!already_parsed;
	    (*i
	      let te =
	      {
	      type_entry_name = base;
	      type_entry_fullfilename = fullname;
	      type_entry_read = false;
	      type_entry_is_interface = false;
	      type_entry_extends = [];
	      type_entry_implements = [];
	      type_entry_contents = [];
	      } 
	      in
	      do_read_dir d dirh ((base, Type_entry te)::env)
	      i*)
	    do_read_dir d dirh env
	| _ -> do_read_dir d dirh env
      else
	do_read_dir d dirh env
    else
      do_read_dir d dirh env
  with
    End_of_file -> Unix.closedir dirh; env
;;

let read_directory d env = 
  (* read directory contents. for each sub-directory found, add it in env as a empty package. *) 
  let dirh = 
      try
	(* could be replaced by Sys.readdir in ocaml 3.07 *)
	Unix.opendir d 
      with
	  (* Sys.Sys_error _ -> ... in 3.07 *)
	  Unix.Unix_error _ ->
	    Format.eprintf "Cannot open directory %s@." d;
	    exit 1
  in
    do_read_dir d dirh env 


 let initial_env () =
  let i = !Config.input_path in
  let env = List.fold_right read_directory i [] in
  let name = !Config.source_package in
    Format.fprintf Config.log "Using source package %s.@." name;
     let pack =
       {
	  package_entry_name = name;
	  package_entry_directories = [name];
	  package_entry_contents_read = true;
	  package_entry_contents = read_directory name [];
       } 
     in 
       if !Config.verbose >= 1 then
	 Format.printf "Adding source_package %s in environment@."
	   name;
       ((name, Package_entry pack)::env)
;;


(***********************


 Phase 1

*****************************)


let find_package_contents env p =
  let c = List.assoc p env in
  match c with
  | Package_entry ce -> 
      if ce.package_entry_contents_read
      then ce.package_entry_contents
      else
	begin
	  if !Config.verbose >= 1 then
	    Format.printf "reading directory %a@." 
	      (Format_misc.fprintf_list (fun form s -> Format.fprintf form "%s@." s)) 
	      ce.package_entry_directories;
	  let new_contents =
	    List.fold_right read_directory ce.package_entry_directories []
	  in
	  ce.package_entry_contents <- new_contents;
	  ce.package_entry_contents_read <- true;
	  (*i
	  Format.printf "****** env after read_directory *******@.";
	  printf_java_context env;
	  Format.printf "@.****** end java context *******@.";
  i*)
	  new_contents
	end
	    
  | _ -> assert false
;;


let rec change_package_contents env p cont =
  match env with
    | [] -> assert false
    | (pa,Package_entry c)::envr ->
        if p=pa then c.package_entry_contents <- cont
        else 
          change_package_contents envr p cont
    | e::envr ->
        change_package_contents envr p cont
;;


let rec add_in_package_rec env package contents =
  match package with
    | [] -> contents@env
    | (_,p)::pl ->      
        try
          let package_contents = find_package_contents env p in
          let new_contents = add_in_package_rec package_contents pl contents in
          change_package_contents env p new_contents;
          env
        with
            Not_found ->
	      assert false
	      (*i
              let new_contents = add_in_package_rec [] pl contents in
              (p,Package_entry {
               package_entry_name = snd id;
	       package_entry_directory = "./" ^ (snd id);
	       package_entry_contents_read = true;
               package_entry_contents = new_contents;
               })::env
	      i*)
;;


let add_in_package env package contents =
  add_in_package_rec env (List.rev package) contents
;;


(* import declarations *)

let rec import_package_rec env p =
  match p with
    | [] -> env
    | (loc,p)::pl -> 
        let e = import_package_rec env pl in
        (*i
          Format.printf "looking %s package contents@." p;
          i*)
        try
          find_package_contents e p
        with
            Not_found ->
              begin
                report_typing_error loc
                  (sprintf "package %s not found" p);
                []
              end  
;;

let import_package env p =
 (*i
  Format.printf "Importing package %s.@." (snd (List.hd p));
 i*)
  let contents = import_package_rec env p in
  List.filter
    (fun (_,e) ->
      match e with
      |	Class_entry _ 
      | Interface_entry _ -> true
      |	Package_entry _ -> false
      | Logic_type_entry _ -> true
      | Logic_entry _ -> true
      | _ -> assert false)
    contents
;;

let object_class_loaded = ref false

let object_class = 
  ref 
    { 
      class_entry_package = "";
      class_entry_name = "<uninitialized Object class>";
      class_entry_extends = None;
      class_entry_implements = [];
      class_entry_contents = [];
      class_entry_constructors = [];
      class_entry_invariant_effects = Java_context.default_effects;
      class_entry_static_invariant_effects = Java_context.default_effects;
      class_entry_invariant_args = [];
      class_entry_static_invariant_args = [];
      class_entry_restore_invariant_args = [];
      class_entry_representation_invariant_effects = Java_context.default_effects;
    }
    
let build_env_after_imports all env comp_unit =
  let env_after_java_lang =
    if !Config.nojavalang then env
    else 
      let a = 
        import_package env 
          [Location.dummy,"lang"; Location.dummy, "java"]
      in
      if all && not !object_class_loaded
      then 
	begin
	  object_class_loaded := true;
	  try
	    match (List.assoc "Object" a) with
	    | Class_entry ce -> object_class := ce
	    | _ -> 
		report_typing_error Location.dummy "Oops! 'Object' is not a class in java.lang package! What have you done?"
	with
	    Not_found -> report_typing_error Location.dummy "Class 'Object' not found in java.lang package!"
	end;
      a@env
  in
  let env_after_imports =
    List.fold_left  
      (* and not fold_right because of order of import declarations matters *)
      (fun env import_decl ->
         match import_decl with
         | Import_package(qi) ->
             (import_package env qi)@env
         | Import_class_or_interface((loc,idclass)::qi) ->
             let a = import_package env qi in          
             begin
               try
		 if all then
                   let c = List.assoc idclass a in
                   (idclass,c)::env
		 else env
               with
                 Not_found ->
                   report_typing_error loc
                     (sprintf "Class or interface %s not found in package" idclass);
                   env
             end
         | Import_class_or_interface([]) ->
             assert false
      ) 
      env_after_java_lang
      comp_unit.cu_imports
  in
  let package_name = comp_unit.cu_package in
  let env_after_imports = 
    if package_name = [] then env_after_imports else
      (import_package env package_name)@env_after_imports
  in
  (*i
    Format.printf "****** java context after imports *******@.";
    printf_java_context env_after_imports;
    Format.printf "@.****** end java context *******@.";
  i*)
  env_after_imports
;;

let rec check_package_name_coherence dirname package_name =
  match package_name with
    | [] -> dirname = Filename.current_dir_name or
		      List.mem dirname !Config.input_path
    | (_,p)::pl ->
        let (d,f) = (Filename.dirname dirname,Filename.basename dirname) in
        if f=p then check_package_name_coherence d pl
        else false
;;


  
let phase1 env filename comp_unit =
  let package_name = comp_unit.cu_package in
  if not 
    (check_package_name_coherence 
       (Filename.dirname filename) package_name)
  then
    report_typing_error (Location.file_start filename)
      "Wrong package declaration";               
  let _ = build_env_after_imports false env comp_unit in
  let (new_type_decls,contents,class_or_interface_found) =
    List.fold_left
      (fun (decl_acc,env_acc,b) (doc_comment,typedecl) ->
         match typedecl with
           | Class class_decl ->
               let class_name = snd class_decl.class_name in
	       let ce = {
                 class_entry_package = List.fold_right (fun (_, id) s -> id ^ "." ^ s) package_name "";
		 class_entry_name = snd class_decl.class_name;
		 class_entry_extends = None;
		 class_entry_implements = [];	
		 class_entry_contents = [];
		 class_entry_constructors = [];
		 class_entry_invariant_effects = default_effects;
		 class_entry_static_invariant_effects = default_effects;
		 class_entry_invariant_args = [];
		 class_entry_static_invariant_args = [];
		 class_entry_restore_invariant_args = [];
		 class_entry_representation_invariant_effects = default_effects;
               }
	       in
               ((doc_comment,Typed_class(ce,class_decl))::decl_acc,
		(class_name, Class_entry ce)::env_acc,true)
           | Interface interface_decl -> 
               let interface_name = snd interface_decl.interface_name in
	       let ie = {
		 interface_entry_name = snd interface_decl.interface_name;
		 interface_entry_extends = [];
		 interface_entry_contents = [];
               }
	       in
               ((doc_comment,Typed_interface(ie,interface_decl))::decl_acc,
		(interface_name, Interface_entry ie)::env_acc,true)
	   | Axiom(id,e) ->
	       ((doc_comment,Axiom(id,e))::decl_acc,env_acc,b)
	   | Logic_typedecl(((_, t) as id), vars) ->
	       ((doc_comment, Logic_typedecl(id, vars))::decl_acc,
		(t,
		 Logic_type_entry
		   ({ logic_type_entry_name = t; 
		      logic_type_entry_vars = List.map snd vars }))::env_acc, 
		b)
	   | Logic_reads(id,t,p,r) ->
	       ((doc_comment,Logic_reads(id,t,p,r))::decl_acc,env_acc,b)
	   | Typed_logic_reads _ -> assert false
	   | Logic_def(id,t,p,e) ->
	       ((doc_comment,Logic_def(id,t,p,e))::decl_acc,env_acc,b)
	   | Typed_logic_def _ -> assert false
	   | Typed_class _ -> assert false
	   | Typed_interface _ -> assert false)
      ([],[],false)
      comp_unit.cu_type_decls
  in
  ({ comp_unit with cu_type_decls = new_type_decls },
   add_in_package env package_name contents)
;;

let additional_files_to_type = ref [];;
  
let rec phase1_imported_files env =
  if !additional_files_to_parse <> [] 
  then
    let l = !additional_files_to_parse in
    additional_files_to_parse := [];
    let new_env =
      List.fold_left
	(fun env f -> 
	   (*i*)
	   if !Config.verbose >=2 then
	     Format.printf "Parsing imported file %s.@." f;	     
	   let a = Java_syntax.parse f in
	   let (cu,new_env) = phase1 env f a in
	   additional_files_to_type :=
	   (f,cu):: !additional_files_to_type;
	   new_env)
	env
	l
    in
    phase1_imported_files new_env
  else
    env
;;
    


(*************************************

  phase 2

**************************************)


let is_static modifiers = List.mem `STATIC modifiers;;

let is_pure modifiers = List.mem `PURE modifiers;;

let is_final modifiers = List.mem `FINAL modifiers;;

let is_ghost modifiers = List.mem `GHOST modifiers;;

let is_model modifiers = List.mem `MODEL modifiers;;

let get_access loc modifiers =
  let l =
    List.filter
      (fun m -> m=`PRIVATE or m=`PROTECTED or m=`PUBLIC) 
      modifiers
  in
  match l with
    | [] -> `NONE
    | [`PRIVATE | `PROTECTED | `PUBLIC as m] -> m
    | _ ->
	report_typing_error loc
	  "too many accessibility modifiers";
	`NONE
;;
	  


exception Qualif_not_found of (Location.t * string);;

let rec assoc_filter filter id env =
  match env with
    | [] -> raise Not_found
    | (x,a)::r ->
	if x=id then
	  if filter a 
          then a 
          else assoc_filter filter id r
	else assoc_filter filter id r
;;

let any_filter a = true

let container_filter a =
  match a with
    | Package_entry _ 
    | Class_entry _ | Interface_entry _
    | Instance_variable_entry _ 
    | Local_variable_entry _ -> true
    | Logic_entry _
    | Logic_type_entry _
    | Method_entry _
    | Constructor_entry _
    | Constant_entry _ -> false
;;

let type_filter a =
  match a with
    | Class_entry _ | Interface_entry _
    | Logic_type_entry _ -> true
    | Logic_entry _
    | Local_variable_entry _
    | Method_entry _
    | Constructor_entry _
    | Instance_variable_entry _
    | Package_entry _
    | Constant_entry _ -> false
;;

let rec get_inherited_fields acc env ce =
  let acc =
    match ce.class_entry_extends with
      | None -> acc
      | Some cee ->
          List.fold_left
          (fun acc (_, entry) ->
             match entry with
               | Instance_variable_entry ive ->
                   begin
                     List.append 
                     (get_inherited_fields [] env cee)
                     (match ive.instance_variable_entry_accessibility with
                       | `PRIVATE -> acc
		       | `PUBLIC | `PROTECTED -> (ive.instance_variable_entry_name, Instance_variable_entry ive)::acc
                       | `NONE ->
                           if ce.class_entry_package = cee.class_entry_package
                           then (ive.instance_variable_entry_name, Instance_variable_entry ive)::acc
                           else acc)
                   end
               | _ -> acc)
          acc
          cee.class_entry_contents
  in
  (* TO DO : interfaces implemented by ce *)
  acc
;;

let rec get_name_rec filter
  (env : Java_context_types.t) (qi : qualified_ident) : java_env_entry =
  match qi with
    | [] -> assert false
    | [(loc,id)] ->
        begin
          try
            assoc_filter filter id env
          with Not_found -> raise (Qualif_not_found(loc,id))
        end
    | (loc,id)::q ->
        let e = match get_name_rec container_filter env q with
          | Package_entry pe -> pe.package_entry_contents
          | Class_entry te -> te.class_entry_contents
          | Interface_entry te -> te.interface_entry_contents
          | Instance_variable_entry ive ->
              begin
                match ive.instance_variable_entry_type with
                 | Class_type ce -> ce.class_entry_contents
                 | Interface_type ie -> ie.interface_entry_contents
                 | _ -> assert false (* TODO ? *)
              end
          | Local_variable_entry lve ->
              begin
                match lve.local_variable_entry_type with
                  | Class_type ce -> List.append (get_inherited_fields [] env ce) ce.class_entry_contents
                  | Interface_type te -> te.interface_entry_contents
                  | _ -> assert false (* TODO ? *)
              end
          | _ -> assert false (* TODO ? *)
        in 
        try
          assoc_filter filter id e
        with Not_found -> raise (Qualif_not_found(loc,id))
;;

let logic_functions = Hashtbl.create 17;;

let get_name (env : Java_context_types.t) (qi : qualified_ident) : java_env_entry =
  try
    get_name_rec any_filter env qi 
  with
      Qualif_not_found(loc,id) ->
(*	try
	  match qi with
	      [(_,id2)] ->
		let le = Hashtbl.find logic_functions id2 in
		(le.logic_entry_return_type, 
		 mk_expr_no_loc (Logic_call(le,[])))
	    | _ -> raise Not_found
	with Not_found ->
*)        begin
            report_typing_error  loc
              (sprintf "Unknown identifier %s" id);
            raise Exit;
          end
;;
        
let get_type_name 
  (env : Java_context_types.t) (qi : qualified_ident) : java_env_entry =
  try
    get_name_rec type_filter env qi 
  with
      Qualif_not_found(loc,id) ->
        begin
          report_typing_error  loc
            (sprintf "Unknown type identifier %s" id);
          raise Exit;
        end


let rec get_type (env : Java_context_types.t) var_type =
  match var_type with
    | Base_type s ->
        begin
          match s with 
            | "byte" -> Byte_type
            | "short" -> Short_type
            | "int" | "long" | "char" -> Integer_type
            | "float" | "double" -> Float_type
	    | "real" -> Real_type
            | "boolean" -> Boolean_type
            | _ -> failwith ("unknown base type : "^s)
        end
    | Type_name(qi, vars) -> 
        begin
          match get_type_name env qi with
            | Class_entry te -> Class_type te
            | Interface_entry te -> Interface_type te
	    | Logic_type_entry lte -> 
		(* TO DO : tous les cas selon vars *)
		Java_context_types.Logic_type
		  (lte.logic_type_entry_name,
		   List.map 
		     (fun v -> Type_var(v)) 
		     lte.logic_type_entry_vars)
            | t ->
                begin
                  match qi with
                    | [] -> assert false
                    | (loc,id)::_ ->
                        report_typing_error  loc
                        (Format.fprintf 
			   Format.str_formatter
			   "%s has type %a, not a class or an interface" 
			   id Java_context.fprintf_java_env_entry t; 
			 Format.flush_str_formatter());
                        raise Exit;
                end
        end
    | Array_type_expr(t) ->
        Array_type(get_type env t)
    | Type_var_name(id) -> Type_var(snd id)
;;


let rec get_var_id_and_type var_type var_id =
  match var_id with
    | Simple_id (loc,id) -> (loc,id,var_type)
    | Array_id var_id -> 
        let (loc,id,t) = get_var_id_and_type var_type var_id in
        (loc,id,Array_type(t))
;;
      
let get_var_decl var_type var_decl =
  let (loc,id,t) = get_var_id_and_type var_type var_decl.variable_id in
  (loc,id,t, var_decl.variable_initializer)
;;

let rec get_parameter class_env p =
  match p with
    | Simple_parameter(t,(loc,id)) ->
        (id,get_type class_env t)
    | Array_parameter(p) ->
        let (id,t) = get_parameter class_env p in
        (id,Array_type(t))
;;

let get_meth_decl class_env class_or_interface_entry access is_static is_pure return_type m_decl m_spec =
  match m_decl with
    | Simple_method_declarator((loc,id),pl) ->
        let params = List.map (get_parameter class_env) pl in   
        let param_env =
          List.map (fun (id,t) ->           
                      (id, Local_variable_entry 
			 { 
                           local_variable_entry_name = id;
                           local_variable_entry_trans_name = id;
                           local_variable_entry_is_ref =
			     if !Config.use_value_model then
			       (not is_pure) && (is_ref_type t)
                               (* TODO : flase unless specified modifiable *)
			     else
			       false;
                           local_variable_entry_type = t; 
			 }))
            params 
        in
        let reads_alloc_table =
          match return_type with
            | Class_type _ | Interface_type _ | Array_type _ -> true
            | _ -> false
        in
	let routine_entry =
	  { Java_context.default_routine with
	      routine_entry_parameters = param_env;
	      routine_entry_effects = 
	      { default_effects with reads_alloc = not is_static; reads_alloc_table = reads_alloc_table }; }
 	in
        let sep_spec =
          match m_spec with
            | Non_extending [sc] ->
                begin 
                match sc with
                  | Behavior(_, gsc) -> 
                      begin
                      match gsc.generic_spec_body with
                        | Nested_spec_body gscl -> List.map (fun gsc -> gsc.generic_spec_label) gscl
                        | _ -> []
                      end
                  | Normal_behavior(_, nsc) -> 
                      begin
                      match nsc.normal_spec_body with
                        | Nested_normal_spec_body nscl -> List.map (fun nsc -> nsc.normal_spec_label) nscl
                        | _ -> []
                      end
                  | _ -> [] (* TO DO: Exceptional_behavior *) 
                end
            | _ -> []
        in
        (id, 
	 { method_entry_name = id;
           method_entry_trans_name = id;
	   method_entry_accessibility = access;
           method_entry_is_static = is_static;
           method_entry_is_pure = is_pure;
           method_entry_return_type = return_type;
	   method_entry_class_or_interface = class_or_interface_entry;
	   method_entry_is_predicate = false;
	   method_entry_routine = routine_entry;
           method_entry_sep_specs = sep_spec;
	   method_entry_graph = [] })
    | Array_method_declarator(md) ->
        failwith "arrays unsupported in Typing.get_meth_decls"
;;

let get_constr_decl class_env class_entry c_decl =
  let params = 
    List.map (get_parameter class_env) c_decl.constr_parameters  
  and (loc,id) = c_decl.constr_name
  in
  let param_env =
    List.map (fun (id,t) ->
                (id, Local_variable_entry { 
                   local_variable_entry_name = id;
                   local_variable_entry_trans_name = id;
                   local_variable_entry_is_ref = false (* TODO : correct ? *);
                   local_variable_entry_type = t; }))
      params 
  in
  let routine_entry =
    { Java_context.default_routine with
      routine_entry_parameters = param_env;
	routine_entry_effects = 
    {default_effects with reads_alloc = true};
}
  in
   (id, 
    { constructor_entry_trans_name = id ^ "_" ^ id;
      constructor_entry_class = class_entry;
      constructor_entry_routine = routine_entry; 
      constructor_entry_graph = [] })
;;

let get_class_jml_decls class_env class_entry doc s =
  match s with
  | Jml_variable_declaration(d) ->
      begin
        try
          let var_type = get_type class_env d.jml_variable_type 
          and accessibility = get_access Location.dummy d.jml_variable_modifiers
          and is_static = is_static d.jml_variable_modifiers
          and is_final = is_final d.jml_variable_modifiers
          and is_ghost = is_ghost d.jml_variable_modifiers
          and is_model = is_model d.jml_variable_modifiers
          in
          List.map
            (fun vd ->
               let (loc,name,t) = get_var_id_and_type var_type vd in
	       let tag = new_tag () in
	       Format.fprintf Config.log "instance variable %s as tag %d@."
		 name tag;
               let d = 
                 { instance_variable_entry_tag = tag;
		   instance_variable_entry_name = name;
		   instance_variable_entry_trans_name = name;
                   instance_variable_entry_accessibility = accessibility;
		   instance_variable_entry_is_static = is_static;
                   instance_variable_entry_is_final = is_final;
		   instance_variable_entry_class_or_interface = Class_entry(class_entry);
                   instance_variable_entry_type = t; 
		   instance_variable_entry_is_model = is_model; 
		   instance_variable_entry_is_ghost = is_ghost; } 
               in
               class_entry.class_entry_contents <- 
               (name, Instance_variable_entry d)::
               class_entry.class_entry_contents;
               (doc,Typed_variable(d,None)))
            d.jml_variable_decls
        with Exit -> [(doc,Specification([s]))]
      end
  | _ -> [(doc,Specification([s]))]
;;

let get_interface_jml_decls interface_env interface_entry doc s =
  match s with
  | Jml_variable_declaration(d) ->
      begin
        try
          let var_type = get_type interface_env d.jml_variable_type 
          and accessibility = get_access Location.dummy d.jml_variable_modifiers
          and is_static = is_static d.jml_variable_modifiers
          and is_final = is_final d.jml_variable_modifiers
          and is_ghost = is_ghost d.jml_variable_modifiers
          and is_model = is_model d.jml_variable_modifiers
          in
          List.map
            (fun vd ->
               let (loc,name,t) = get_var_id_and_type var_type vd in
	       let tag = new_tag () in
	       Format.fprintf Config.log "instance variable %s as tag %d@."
		 name tag;
               let d = 
                 { instance_variable_entry_tag = tag;
		   instance_variable_entry_name = name;
		   instance_variable_entry_trans_name = name;
                   instance_variable_entry_accessibility = accessibility;
		   instance_variable_entry_is_static = is_static;
                   instance_variable_entry_is_final = is_final;
		   instance_variable_entry_class_or_interface = Interface_entry(interface_entry);
                   instance_variable_entry_type = t; 
		   instance_variable_entry_is_model = is_model; 
		   instance_variable_entry_is_ghost = is_ghost; } 
               in
               interface_entry.interface_entry_contents <- 
               (name, Instance_variable_entry d)::
               interface_entry.interface_entry_contents;
               (doc, Typed_model_variable(d)))
            d.jml_variable_decls
        with Exit -> [(doc,Interface_specification([s]))]
      end
  | _ -> [(doc,Interface_specification([s]))]
;;

let rec add_method_in_env id d env =
  match env with
    | [] -> [(id,Method_entry [d])]
    | (x,Method_entry l)::r when x=id ->
	(id,Method_entry (d::l))::r
    | x::r -> x::(add_method_in_env id d r)
;;

let get_class_decls class_env class_entry doc field =
  match field with
    | Method(m_decl,block_opt) ->
      begin
        try
          let return_type = 
            match m_decl.method_return_type with
              | None -> Null_type
              | Some t -> get_type class_env t
          in
          let (id,d) = 
            get_meth_decl 
              class_env (Class_entry class_entry)
	      (get_access Location.dummy m_decl.method_modifiers)
              (is_static m_decl.method_modifiers) 
              (is_pure m_decl.method_modifiers) 
              return_type m_decl.method_declarator
              m_decl.method_specification 
          in 
          class_entry.class_entry_contents <- 
          add_method_in_env id d class_entry.class_entry_contents;
          [(doc,Typed_method(d,m_decl.method_specification,block_opt))]
        with Exit -> [(doc,field)]
      end
  | Constructor(const_decl) ->
      begin
        try
          let (id,d) = 
            get_constr_decl class_env class_entry const_decl
          and (ice,b) = const_decl.constr_body
          in 
          class_entry.class_entry_constructors <- 
          d::class_entry.class_entry_constructors;
          [(doc,Typed_constructor(d,const_decl.constr_specification,ice,b))]
        with Exit -> [(doc,field)]
      end
  | Variable(var_decl) ->
      begin
        try
          let var_type = get_type class_env var_decl.variable_type 
          and accessibility = get_access Location.dummy var_decl.variable_modifiers
          and is_static = is_static var_decl.variable_modifiers
          and is_final = is_final var_decl.variable_modifiers
          and is_ghost = is_ghost var_decl.variable_modifiers 
          and is_model = is_model var_decl.variable_modifiers 
          in
          List.map
            (fun vd ->
               let (loc,name,t,init) = get_var_decl var_type vd in
	       let tag = new_tag () in
	       Format.fprintf Config.log "instance variable %s as tag %d@."
		 name tag ;
               let d = 
                 { instance_variable_entry_tag = tag;
		   instance_variable_entry_name = name;
		   instance_variable_entry_trans_name = 
		     if is_static then
		       class_entry.class_entry_name ^ "_" ^ name
		     else name ;
                   instance_variable_entry_accessibility = accessibility;
                   instance_variable_entry_is_static = is_static;
                   instance_variable_entry_is_final = is_final;
		   instance_variable_entry_class_or_interface = Class_entry(class_entry);
                   instance_variable_entry_type = t; 
                   instance_variable_entry_is_model = is_model ;
                   instance_variable_entry_is_ghost = is_ghost; } 
               in
               class_entry.class_entry_contents <- 
               (name, Instance_variable_entry d)::
               class_entry.class_entry_contents;
               (doc,Typed_variable(d,init)))
            var_decl.variable_decls
        with Exit -> [(doc,field)]
      end
  | Static_initializer(block) ->
      Format.printf 
        "Warning: static initializers unsupported in Typing.get_class_decls@.";
      [(doc,field)]
  | Specification(jml_spec) ->
      List.fold_left 
	(fun acc s -> 
	  (get_class_jml_decls class_env class_entry doc s)@acc)
	[] jml_spec
  | Typed_method(_) -> assert false (*r Should never occur *)
  | Typed_constructor(_) -> assert false (*r Should never occur *)
  | Typed_variable(_) -> assert false (*r Should never occur *)
;; 

let get_interface_decls interface_env interface_entry doc member =
  match member with
  | Constant_declaration(var_decl) ->
      begin
        try
          let var_type = get_type interface_env var_decl.variable_type 
          and _is_static = is_static var_decl.variable_modifiers
          in
	  (*
          if not is_static then
            begin
              report_typing_error  Location.dummy 
                "this variable should be declared static";
              raise Exit
            end
          else
          *)
	  let _is_final = is_final var_decl.variable_modifiers in
          List.map
            (fun vd ->
               let (loc,name,t,init) = get_var_decl var_type vd in
               let d = 
                 { constant_entry_name = name;
		   constant_entry_interface = interface_entry;
                   constant_entry_type = t; }  
               in
               interface_entry.interface_entry_contents <- 
               (name, Constant_entry d)::
                 interface_entry.interface_entry_contents;
               (doc,Typed_constant_declaration(d,init)))
            var_decl.variable_decls
        with Exit -> [(doc,member)]
      end
  | Method_header(m_decl) ->
      begin
        try
          let return_type = 
            match m_decl.method_return_type with
              | None -> Null_type
              | Some t -> get_type interface_env t
          in
          let (id,d) = 
            get_meth_decl 
              interface_env (Interface_entry interface_entry)
	      (get_access Location.dummy m_decl.method_modifiers)
              (is_static m_decl.method_modifiers) 
              (is_pure m_decl.method_modifiers) 
              return_type m_decl.method_declarator
              m_decl.method_specification 
          in 
          interface_entry.interface_entry_contents <- 
          add_method_in_env id d interface_entry.interface_entry_contents;
          [(doc,Typed_method_header(d,m_decl.method_specification))]
        with Exit -> [(doc,member)]
      end
  | Typed_constant_declaration(_) -> assert false (*r Should never occur *)
  | Typed_method_header(_) -> assert false (*r Should never occur *)
  | Model_variable(var_decl) ->
      begin
        try
          let var_type = get_type interface_env var_decl.variable_type 
          and accessibility = get_access Location.dummy var_decl.variable_modifiers
          and is_static = is_static var_decl.variable_modifiers
          and is_final = is_final var_decl.variable_modifiers
          and is_ghost = is_ghost var_decl.variable_modifiers 
          and is_model = is_model var_decl.variable_modifiers
          in
	  if not is_model 
	  then
	    begin
              report_typing_error Location.dummy 
                "only model variables are allowed in Java interface declaration";
              raise Exit
            end
          else
          List.map
            (fun vd ->
               let (loc,name,t,_) = get_var_decl var_type vd in
	       let tag = new_tag () in
	       Format.fprintf Config.log "instance variable %s as tag %d@."
		 name tag ;
               let d = 
                 { instance_variable_entry_tag = tag;
		   instance_variable_entry_name = name;
		   instance_variable_entry_trans_name = 
		     if is_static then
		       interface_entry.interface_entry_name ^ "_" ^ name
		     else name ;
                   instance_variable_entry_accessibility = accessibility;
                   instance_variable_entry_is_static = is_static;
                   instance_variable_entry_is_final = is_final;
		   instance_variable_entry_class_or_interface = Interface_entry(interface_entry);
                   instance_variable_entry_type = t; 
                   instance_variable_entry_is_model = is_model;  
                   instance_variable_entry_is_ghost = is_ghost; } 
               in
               interface_entry.interface_entry_contents <- 
               (name, Instance_variable_entry d)::
               interface_entry.interface_entry_contents;
               (doc,Typed_model_variable(d)))
            var_decl.variable_decls
        with Exit -> [(doc, member)]
      end
  | Typed_model_variable(_) -> assert false (*r Should never occur *)
  | Interface_specification(jml_spec) ->
      List.fold_left 
	(fun acc s -> 
	  (get_interface_jml_decls interface_env interface_entry doc s)@acc)
	[] jml_spec
;; 




let phase2 global_env filename comp_unit =
  let env_after_imports =
    build_env_after_imports true global_env comp_unit
  in
  let new_typedecl = 
    List.map
      (fun (doc_comment,typedecl) ->
         let new_typedecl =
           match typedecl with
	   | Class _ -> assert false
	   | Interface _ -> assert false
	   | Axiom _ -> typedecl
	   | Logic_typedecl _ -> typedecl
	   | Logic_reads(id,t,p,r) -> 
               let params = List.map (get_parameter env_after_imports) p in   
               let param_env =
		 List.map (fun (id,t) ->           
			     (id, Local_variable_entry 
				{ 
				  local_variable_entry_name = id;
				  local_variable_entry_trans_name = id;
				  local_variable_entry_is_ref = false;
				  local_variable_entry_type = t; 
				}))
		   params 
               in
	       let ty = match t with
		 | None -> Prop_type
		 | Some t -> get_type env_after_imports t
	       in
	       let le =
		 {
		   logic_entry_name = snd id;
		   logic_entry_return_type = ty;
		   logic_entry_effects = default_effects;
		   logic_entry_parameters = param_env;
		 }
	       in
	       Hashtbl.add logic_functions (snd id) le;
	       Typed_logic_reads(le,r)
	   | Logic_def(id,t,p,e) -> 
               let params = List.map (get_parameter env_after_imports) p in   
               let param_env =
		 List.map (fun (id,t) ->           
			     (id, Local_variable_entry 
				{ 
				  local_variable_entry_name = id;
				  local_variable_entry_trans_name = id;
				  local_variable_entry_is_ref = false;
				  local_variable_entry_type = t; 
				}))
		   params 
               in
	       let ty = match t with
		 | None -> Prop_type
		 | Some t -> get_type env_after_imports t
	       in
	       let le =
		 {
		   logic_entry_name = snd id;
		   logic_entry_return_type = ty;
		   logic_entry_effects = default_effects;
		   logic_entry_parameters = param_env;
		 }
	       in
	       Hashtbl.add logic_functions (snd id) le;
	       Typed_logic_def(le,e)
	   | Typed_logic_reads _ -> assert false
	   | Typed_logic_def _ -> assert false
           | Typed_class(class_entry, class_decl) ->
		 begin
		   match class_decl.class_extends with
		   | None -> 
		       if class_entry != !object_class
		       then 
			 class_entry.class_entry_extends <- 
			 Some !object_class
		   | Some qi ->
		       try
			 match get_name env_after_imports qi with
			 | Class_entry te ->
			     class_entry.class_entry_extends <- Some te
			 | _ ->              
			     let (l,id) = List.hd qi in
			     report_typing_error  
                               l (sprintf "%s is not a class" id);
		       with
			 Exit -> ()
		 end;	         
		 class_entry.class_entry_implements <- 
		   List.fold_left 
		     (fun acc qi ->
		       try
			 match get_name env_after_imports qi with
			 | Interface_entry te ->
			     te::acc
			 | _ ->              
			     let (l,id) = List.hd qi in
			     report_typing_error  
                               l (sprintf "%s is not an interface" id);
			     acc
		       with Exit -> acc)
		     []
		     class_decl.class_implements;
                 let new_fields = 
                   List.fold_left
                     (fun acc (doc,c) -> 
                       (get_class_decls 
                          env_after_imports class_entry doc c)
                       @acc)
                     []
                     class_decl.class_fields
                 in
                 Typed_class
		   (class_entry,
		    { class_decl with class_fields = new_fields})
             | Typed_interface(interface_entry, interface_decl) -> 
		 interface_entry.interface_entry_extends <- 
		   List.fold_left 
		     (fun acc qi ->
		       try
			 match get_name env_after_imports qi with
			 | Interface_entry te ->
			     te::acc
			 | _ ->              
			     let (l,id) = List.hd qi in
			     report_typing_error  
                               l (sprintf "%s is not an interface" id);
			     acc
		       with Exit -> acc)
		     []
		     interface_decl.interface_extends;
                 let new_members =
                   List.fold_left
                     (fun acc (doc,c) -> 
                       (get_interface_decls 
                          env_after_imports interface_entry doc c)@acc)
                     []
                     interface_decl.interface_members
                 in
                 Typed_interface
                   (interface_entry,
		    { interface_decl with interface_members = new_members})
         in (doc_comment,new_typedecl))
      comp_unit.cu_type_decls
  in
  { comp_unit with cu_type_decls = new_typedecl }
;;




let rec constructor_overloading env =
  List.iter
    (fun (_,entry) -> match entry with
       | Package_entry pe ->
	   constructor_overloading pe.package_entry_contents
       | Class_entry te ->
	   begin
	     match te.class_entry_constructors with
	       | [] -> 
		   Format.fprintf Config.log 
		     "Adding default constructor to class %s@."
		     te.class_entry_name;
		   let routine_entry = Java_context.default_routine in
 		   let c =
		     { constructor_entry_trans_name = 
		       te.class_entry_name ^ "_" ^ te.class_entry_name;
		       constructor_entry_class = te ;
		       constructor_entry_routine = routine_entry;
		       constructor_entry_graph = [] }
		   in
		   te.class_entry_constructors <- [ c ]
	       | [_] -> ()
	       | l -> 
		   let n =
		     List.fold_right 
		       (fun c acc ->
			  let n = acc + 1 in
			  c.constructor_entry_trans_name <- 
			  c.constructor_entry_trans_name ^ (string_of_int n);
			  n)
		       l 0
		   in
		   Format.fprintf Config.log 
		     "Class %s has %d constructors.@."
		     te.class_entry_name n;		   
	   end
       | _ -> ())
    env

(*************************************

  phase 3

**************************************)


let type_lit l =
  match l with
    | Int(n) -> Integer_type
    | Long(n) -> Integer_type
    | Float _ -> Float_type
    | Bool(b) -> Boolean_type
    | String(s) -> assert false (* Class_type("java.lang.String") *)
    | Char(c) -> Integer_type
    | Null -> Null_type
;;


let bool_op_of_bin_arith_op op =
  match op with 
    | "==" -> "eq_bool"
    | "!=" -> "ne_bool"  
    | "&"  -> "andb"
    | "|"  -> "orb"
    | _ -> assert false
;;

let int_op_of_bin_arith_op op =
  match op with 
    | "+" -> "add_int"
    | "-" -> "sub_int"
    | "*" -> "mul_int"
    | "/" -> "div_int"
    | "%" -> "mod_int"
    | "==" -> "eq_int"
    | "<=" -> "le_int"
    | ">=" -> "ge_int"
    | "<"  -> "lt_int"
    | ">"  -> "gt_int"
    | "!=" -> "ne_int"  
    | "&"  -> "Zland"
    | "|"  -> "Zlor"
    | "^"  -> "Zlxor"
    | ">>" -> "Zasr"
    | "<<" -> "Zlsl"
    | ">>>" -> "Zlsr"
    | _ -> assert false
;;

let float_op_of_bin_arith_op op =
  match op with 
    | "+" -> "add_float"
    | "-" -> "sub_float"
    | "*" -> "mul_float"
    | "/" -> "div_float"
    | "==" -> "eq_float"
    | "<=" -> "le_float"
    | ">=" -> "ge_float"
    | "<"  -> "lt_float"
    | ">"  -> "gt_float"
    | "!=" -> "ne_float"        
    | _ -> assert false
;;

let ref_op_of_bin_arith_op op =
  match op with 
    | "==" -> "eq"
    | "!=" -> "neq"        
    | _ -> assert false
;;

let type_op_of_bin_arith_op op =
  match op with 
    | "==" -> "eq_type"
    | "!=" -> "neq_type"        
    | _ -> assert false
;;

let logic_op_of_bin_arith_op op =
  match op with 
    | "==" -> "eq"
    | "!=" -> "neq"        
    | _ -> assert false
;;

let type_bin_op loc float_of_int op t1 t2 te1 te2 =
  match op with
    | "+" | "-" | "*" | "/" ->
        begin
          match (t1,t2) with
            | (Byte_type, Byte_type)
            | (Byte_type, Short_type)
            | (Byte_type, Integer_type)
            | (Short_type, Byte_type)
            | (Short_type, Short_type)
            | (Short_type, Integer_type)
            | (Integer_type, Byte_type)
            | (Integer_type, Short_type)
            | (Integer_type,Integer_type) -> 
                (Integer_type,te1,int_op_of_bin_arith_op op,te2)
            | (Float_type, Float_type) -> 
                (Float_type,te1,float_op_of_bin_arith_op op,te2)
            | (Integer_type,Float_type) -> 
                (Float_type,
                 (float_of_int te1),float_op_of_bin_arith_op op,te2)
            | (Float_type,Integer_type) -> 
                (Float_type,
                 te1,float_op_of_bin_arith_op op,(float_of_int te2))
            | (Real_type, Real_type) -> 
                (Real_type, te1, float_op_of_bin_arith_op op, te2)
            | (Float_type, Real_type) ->
		(Real_type, te1, float_op_of_bin_arith_op op, te2) (* CORRECT ? *)
            | _,_ -> 
                report_typing_error
                  loc
                  (Format.fprintf 
                     Format.str_formatter 
                     "cannot have type %a and type %a with operator %s" 
                     Java_context.fprintf_java_type t1 
                     Java_context.fprintf_java_type t2 op;
                   Format.flush_str_formatter ());
                (Null_type,te1,op,te2)
        end
    | "&" | "|" | "^" ->
        begin
          match (t1,t2) with
            | (Byte_type, Byte_type)
            | (Byte_type, Short_type)
            | (Byte_type, Integer_type)
            | (Short_type, Byte_type)
            | (Short_type, Short_type)
            | (Short_type, Integer_type)
            | (Integer_type, Byte_type)
            | (Integer_type, Short_type)
            | (Integer_type,Integer_type) -> 
                   (Integer_type,te1,int_op_of_bin_arith_op op,te2)
            | (Boolean_type,Boolean_type) -> 
                   (Boolean_type,te1,bool_op_of_bin_arith_op op,te2)
            | _,_ -> 
                report_typing_error
                  loc
                  (Format.fprintf 
                     Format.str_formatter 
                     "cannot have type %a and type %a with operator %s" 
                     Java_context.fprintf_java_type t1 
                     Java_context.fprintf_java_type t2 op;
                   Format.flush_str_formatter ());
                (Null_type,te1,op,te2)
        end
    | "%" | "<<" | ">>" | ">>>" ->
        begin
          match (t1,t2) with
            | (Byte_type, Byte_type)
            | (Byte_type, Short_type)
            | (Byte_type, Integer_type)
            | (Short_type, Byte_type)
            | (Short_type, Short_type)
            | (Short_type, Integer_type)
            | (Integer_type, Byte_type)
            | (Integer_type, Short_type)
            | (Integer_type,Integer_type) -> 
                   (Integer_type,te1,int_op_of_bin_arith_op op,te2)
            | _,_ -> 
                report_typing_error
                  loc
                  (Format.fprintf 
                     Format.str_formatter 
                     "cannot have type %a and type %a with operator %s" 
                     Java_context.fprintf_java_type t1 
                     Java_context.fprintf_java_type t2 op;
                   Format.flush_str_formatter ());
                (Null_type,te1,op,te2)
        end
    | "<=" | ">=" | "<" | ">" ->
        begin
          match (t1,t2) with
            | (Byte_type, Byte_type)
            | (Byte_type, Short_type)
            | (Byte_type, Integer_type)
            | (Short_type, Byte_type)
            | (Short_type, Short_type)
            | (Short_type, Integer_type)
            | (Integer_type, Byte_type)
            | (Integer_type, Short_type)
            | (Integer_type, Integer_type) -> 
                (Boolean_type,te1,int_op_of_bin_arith_op op,te2)
            | (Float_type,Float_type) -> 
                (Boolean_type,te1,float_op_of_bin_arith_op op,te2)
            | (Integer_type,Float_type) -> 
                (Boolean_type,
                 (float_of_int te1),float_op_of_bin_arith_op op,te2)
            | (Float_type,Integer_type) -> 
                (Boolean_type,
                 te1,float_op_of_bin_arith_op op,(float_of_int te2))
            | _,_ -> 
                report_typing_error
                  loc
                  (Format.fprintf 
                     Format.str_formatter 
                     "cannot compare type %a and type %a with operator %s" 
                     Java_context.fprintf_java_type t1 
                     Java_context.fprintf_java_type t2 op;
                   Format.flush_str_formatter ());
                (Null_type,te1,op,te2)
        end
    | "<:" ->
        begin
          match (t1,t2) with
            | (Type_type,Type_type) 
              -> (Boolean_type, te1,"le_type",te2)
            | _,_ -> 
                report_typing_error
                   loc
                  (Format.fprintf 
                     Format.str_formatter 
                     "cannot compare type %a and type %a with operator %s" 
                     Java_context.fprintf_java_type t1 
                     Java_context.fprintf_java_type t2 op;
                   Format.flush_str_formatter ());
                (Null_type,te1,op,te2)
        end
     | "==" | "!=" ->
        begin
          match (t1,t2) with
            | (Boolean_type,Boolean_type) -> 
                (Boolean_type,te1,bool_op_of_bin_arith_op op,te2)
            | (Byte_type, Byte_type)
            | (Byte_type, Short_type)
            | (Byte_type, Integer_type)
            | (Short_type, Byte_type)
            | (Short_type, Short_type)
            | (Short_type, Integer_type)
            | (Integer_type, Byte_type)
            | (Integer_type, Short_type)
            | (Integer_type,Integer_type) -> 
                (Boolean_type,te1,int_op_of_bin_arith_op op,te2)
            | (Float_type, Float_type) | (Real_type, Real_type) -> 
                (Boolean_type,te1,float_op_of_bin_arith_op op,te2)
            | (Integer_type,Float_type) -> 
                (Boolean_type,
                 (float_of_int te1),float_op_of_bin_arith_op op,te2)
            | (Float_type,Integer_type) -> 
                (Boolean_type,
                 te1,float_op_of_bin_arith_op op,(float_of_int te2))
            | (Real_type, Float_type) ->
		(Boolean_type, te1, float_op_of_bin_arith_op op, te2) (* CORRECT ? *)
            | (Array_type(_),Array_type(_)) 
            | (Class_type(_),Class_type(_)) 
            | (Interface_type(_),Interface_type(_)) 
            | (Array_type(_),Null_type) 
            | (Class_type(_),Null_type) 
            | (Interface_type(_),Null_type) 
            | (Null_type,Array_type(_)) 
            | (Null_type,Class_type(_)) 
            | (Null_type,Interface_type(_)) 
              -> (Boolean_type, te1,ref_op_of_bin_arith_op op,te2)
            | (Type_type, Type_type) 
              -> (Boolean_type, te1, type_op_of_bin_arith_op op, te2)
            | (Logic_type(s1, []), Logic_type(s2, [])) when s1 = s2 -> 
		(Prop_type, te1, logic_op_of_bin_arith_op op, te2)
            | _,_ -> 
                report_typing_error
                   loc
                  (Format.fprintf 
                     Format.str_formatter 
                     "cannot compare type %a and type %a with operator %s" 
                     Java_context.fprintf_java_type t1 
                     Java_context.fprintf_java_type t2 op;
                   Format.flush_str_formatter ());
                (Null_type,te1,op,te2)
        end
    | "&&" | "||" | "==>" | "<==" | "<==>" | "<=!=>" ->
        begin
          match (t1,t2) with
            | (Boolean_type,Boolean_type) -> 
                (Boolean_type,te1,op,te2)
            | ((Prop_type | Boolean_type),(Prop_type | Boolean_type)) -> 
                (Prop_type,te1,op,te2)
            | _,_ -> 
                report_typing_error
                   loc
                  (Format.fprintf 
                     Format.str_formatter 
                     "cannot have type %a and type %a with operator %s" 
                     Java_context.fprintf_java_type t1 
                     Java_context.fprintf_java_type t2 op;
                   Format.flush_str_formatter ());
                (Null_type,te1,op,te2)
        end
    | _ -> 
        report_typing_error
           loc
          (sprintf "unsupported binary operator %s" op);
        (Null_type,te1,op,te2)
;;

let rec is_subclass c1 c2 = 
  if c1==c2 then true
  else if c1 == !object_class then false
  else 
    match c1.class_entry_extends with
      | Some c -> is_subclass c c2
      | None -> assert false (* c2 == !object_class *)
;;

let rec is_subinterface i1 i2 =
  if i1 == i2 
  then true
  else
    match i1.interface_entry_extends with
     | [] -> false
     | l -> List.exists (fun i -> is_subinterface i i2) l
;;

let rec is_subtype t1 t2 =
  match (t1,t2) with
    | (Byte_type, Short_type) 
    | (Byte_type, Integer_type) 
    | (Short_type, Integer_type) 
    | (Null_type, Class_type _)
    | (Null_type, Interface_type _)
    | (Null_type, Array_type _) -> true
    | (Class_type(c1), Class_type(c2)) -> is_subclass c1 c2
    | (Interface_type(ie1), Interface_type(ie2)) -> is_subinterface ie1 ie2
    | (Array_type(t1'), Array_type(t2')) -> is_subtype t1' t2'
    | (Array_type _, Class_type(c)) -> c.class_entry_package = "lang.java." && c.class_entry_name = "Object"
    | (Interface_type _, Class_type c) -> c.class_entry_package = "lang.java." && c.class_entry_name = "Object" 
    | _ -> t1 = t2
;;

let type_assignment_op loc (float_of_int : Ast_types.expr -> Ast_types.expr) 
  op0 t1 t2 te2 =
  if op0 = "="
  then
    match (t1,t2) with
      | (Byte_type, Byte_type) 
      | (Byte_type, Short_type) 
      | (Byte_type, Integer_type) 
      | (Short_type, Byte_type) 
      | (Short_type, Short_type) 
      | (Short_type, Integer_type) 
      | (Integer_type, Byte_type) 
      | (Integer_type, Short_type) 
      | (Integer_type,Integer_type) -> 
          (op0,te2)
      | (Float_type,Float_type) -> 
          (op0,te2)
      | (Float_type,Integer_type) -> 
          (op0,(float_of_int te2))
      | (Boolean_type,Boolean_type) -> 
          (op0,te2)
      | (Class_type c1,Null_type) -> 
	  (op0,te2)
      | (Class_type c1,Class_type c2) -> 
          if is_subclass c2 c1 then (op0,te2) else
            begin
	      report_typing_error
		loc
		(Format.fprintf 
		   Format.str_formatter 
		   "class %s is not a sub-class of %s in assignment" 
		   c1.class_entry_name c2.class_entry_name;
		 Format.flush_str_formatter ());
              (op0,te2)
	    end
      | (Class_type ce, Interface_type ie) -> 
          if ce.class_entry_package = "lang.java." && 
             ce.class_entry_name = "Object" 
          then (op0, te2) 
          else
            begin
	      report_typing_error
		loc
		(Format.fprintf 
		   Format.str_formatter 
		   "cannot have type %s and type %s in assignment" 
		   ce.class_entry_name ie.interface_entry_name;
		 Format.flush_str_formatter ());
              (op0, te2)
	    end
      | (Class_type ce, Array_type jt) -> 
          if ce.class_entry_package = "lang.java." && 
             ce.class_entry_name = "Object" 
          then (op0, te2) 
          else
            begin
	      report_typing_error
		loc
		(Format.fprintf 
		   Format.str_formatter 
		   "cannot have type %s and type %a in assignment" 
		   ce.class_entry_name Java_context.fprintf_java_type jt;
		 Format.flush_str_formatter ());
              (op0, te2)
	    end
      | (Interface_type i, Null_type) -> 
	  (op0, te2)
      | (Interface_type i1, Interface_type i2) -> 
          if is_subinterface i2 i1 
          then (op0, te2)
          else
            begin
	      report_typing_error
		loc
		(Format.fprintf 
		   Format.str_formatter 
		   "interface %s is not a sub-interface of %s in assignment" 
		   i2.interface_entry_name i1.interface_entry_name;
		 Format.flush_str_formatter ());
              (op0, te2)
	    end
      | (Array_type t1, Null_type) -> 
	  (op0,te2)
      | (Array_type t1, Array_type t2) -> 
          if is_subtype t2 t1 then (op0,te2) else
            begin
	      report_typing_error
		loc
		(Format.fprintf 
		   Format.str_formatter 
		   "type %a is not a sub-type of %a in assignment" 
               Java_context.fprintf_java_type t2 
               Java_context.fprintf_java_type t1;
		 Format.flush_str_formatter ());
              (op0,te2)
	    end
      | _,_ -> 
          report_typing_error
            loc
            (Format.fprintf 
               Format.str_formatter 
               "cannot have type %a and type %a in assignment" 
               Java_context.fprintf_java_type t1 
               Java_context.fprintf_java_type t2;
             Format.flush_str_formatter ());
          (op0,te2)
  else
    let op = String.sub op0 0 (String.length op0 - 1) in
    match op with
      | "+" | "-" | "*" | "/" ->
          begin
            match (t1,t2) with
              | (Byte_type, Byte_type)
              | (Byte_type, Short_type)
              | (Byte_type, Integer_type)
              | (Short_type, Byte_type)
              | (Short_type, Short_type)
              | (Short_type, Integer_type)
              | (Integer_type, Byte_type)
              | (Integer_type, Short_type)
              | (Integer_type,Integer_type) -> 
                  (int_op_of_bin_arith_op op,te2)
	      | (Float_type,Float_type) -> 
                  (float_op_of_bin_arith_op op,te2)
              | (Float_type,Integer_type) -> 
                  (float_op_of_bin_arith_op op,(float_of_int te2))
            | _,_ -> 
                report_typing_error
                  loc
                  (Format.fprintf 
                     Format.str_formatter 
                     "cannot have type %a and type %a with operator %s" 
                     Java_context.fprintf_java_type t1 
                     Java_context.fprintf_java_type t2 op0;
                   Format.flush_str_formatter ());
                (op,te2)
        end
    | "%" | "<<" | ">>" | ">>>" ->
        begin
          match (t1,t2) with
            | (Byte_type, Byte_type)
            | (Byte_type, Short_type)
            | (Byte_type, Integer_type)
            | (Short_type, Byte_type)
            | (Short_type, Short_type)
            | (Short_type, Integer_type)
            | (Integer_type, Byte_type)
            | (Integer_type, Short_type)
            | (Integer_type,Integer_type) -> 
                   (int_op_of_bin_arith_op op,te2)
            | _,_ -> 
                report_typing_error
                  loc
                  (Format.fprintf 
                     Format.str_formatter 
                     "cannot have type %a and type %a with operator %s" 
                     Java_context.fprintf_java_type t1 
                     Java_context.fprintf_java_type t2 op0;
                   Format.flush_str_formatter ());
                (op,te2)
        end
    | "&" | "|" ->
        begin
          match (t1,t2) with
            | (Byte_type, Byte_type)
            | (Byte_type, Short_type)
            | (Byte_type, Integer_type)
            | (Short_type, Byte_type)
            | (Short_type, Short_type)
            | (Short_type, Integer_type)
            | (Integer_type, Byte_type)
            | (Integer_type, Short_type)
            | (Integer_type,Integer_type) -> 
                   (int_op_of_bin_arith_op op,te2)
            | (Boolean_type,Boolean_type) -> 
                   (bool_op_of_bin_arith_op op,te2)
            | _,_ -> 
                report_typing_error
                  loc
                  (Format.fprintf 
                     Format.str_formatter 
                     "cannot have type %a and type %a with operator %s" 
                     Java_context.fprintf_java_type t1 
                     Java_context.fprintf_java_type t2 op0;
                   Format.flush_str_formatter ());
                (op,te2)
        end
    | "&&" | "||" ->
        begin
          match (t1,t2) with
            | (Boolean_type,Boolean_type) -> 
                (op,te2)
            | _,_ -> 
                report_typing_error
                   loc
                  (Format.fprintf 
                     Format.str_formatter 
                     "cannot have type %a and type %a with operator %s" 
                     Java_context.fprintf_java_type t1 
                     Java_context.fprintf_java_type t2 op0;
                   Format.flush_str_formatter ());
                (op,te2)
        end
    | _ -> 
        report_typing_error
           loc
          (sprintf "unsupported assignment operator %s" op0);
        (op,te2)
;;

let is_bool_or_prop = function
  | Prop_type | Boolean_type -> true
  | _ -> false
	


let type_un_op loc make_un op t1 te1 =
  match op with
    | "+" | "-" ->
        begin
          match t1 with
            | Integer_type -> 
                 (Integer_type,
                     if op="-" then make_un "neg_int" te1 else te1)     
            | Float_type -> 
                 (Float_type,
                     if op="-" then make_un "neg_float" te1 else te1)   
            | _ -> 
                report_typing_error
                   loc
                  (Format.fprintf 
                     Format.str_formatter 
                     "cannot have type %a with unary operator %s" 
                     Java_context.fprintf_java_type t1 op; 
                   Format.flush_str_formatter ());
                (Null_type,make_un op te1)
        end
    | "!"  ->
        begin
          if is_bool_or_prop t1 then
	    begin
	      (*i
		Format.fprintf Config.log "Typed: !%a@."
		fprintf_expr te1;
		i*)
              (Boolean_type,make_un "!" te1)
	    end
	  else
	    begin
              report_typing_error
                loc
                (Format.fprintf 
                   Format.str_formatter 
                   "cannot have type %a with unary operator %s" 
                   Java_context.fprintf_java_type t1 op; 
                 Format.flush_str_formatter ());
              (Null_type,make_un op te1)
	    end
        end
    | "~"  ->
        begin
          match t1 with
            | Byte_type 
            | Short_type
            | Integer_type -> 
                 (Integer_type,make_un "lcompl" te1)
            | _ -> 
                report_typing_error
                   loc
                  (Format.fprintf 
                     Format.str_formatter 
                     "cannot have type %a with unary operator %s" 
                     Java_context.fprintf_java_type t1 op; 
                   Format.flush_str_formatter ());
                (Null_type,make_un op te1)
        end
    | _ -> 
        report_typing_error
           loc
          (sprintf "unsupported unary operator %s" op);
        (Null_type,make_un op te1)
;;

let type_prepost_op loc make_un op t1 te1 =
  match t1 with
    | Byte_type ->
        (Byte_type, make_un op te1)
    | Short_type ->
        (Short_type, make_un op te1)
    | Integer_type ->
        (Integer_type, make_un op te1)
    | _ -> 
        report_typing_error
        loc
        (Format.fprintf 
           Format.str_formatter 
           "cannot have type %a with pre/post incr/decr op" 
           Java_context.fprintf_java_type t1; 
         Format.flush_str_formatter ());
        (Null_type,make_un op te1)
;;

let rec find_constructor constr_list targs =
  match constr_list with
    | [] -> raise Not_found
    | ce::l ->
	let params = ce.constructor_entry_routine.routine_entry_parameters in
	if List.length params = List.length targs
	then
	  if List.for_all2 is_subtype targs 
	    (List.map (fun (id,t) -> 
			 match t with
			   | Local_variable_entry lve ->
			       lve.local_variable_entry_type
			   | _ -> assert false)
	       params)
	  then ce
	  else find_constructor l targs
	else find_constructor l targs
;;

let rec get_interface_env ie =
  let e =
    List.fold_left
      (fun e i -> 
	 (get_interface_env i)@e)
      []
      ie.interface_entry_extends
  in
    ie.interface_entry_contents@e
;;

let rec get_class_env ce =
  let super_env =
    match ce.class_entry_extends with
      | None -> []
      | Some c ->
	  assert (ce.class_entry_name <> "Object");
	  get_class_env c 
  in
  let implem_env =
    List.fold_left
      (fun e i -> 
	 (get_interface_env i)@e)
      []
      ce.class_entry_implements
  in
    ce.class_entry_contents@super_env@implem_env
;;

(*
let is_static_field (_,f) =
  match f with
    | Method_entry me -> me.method_entry_is_static
    | Constant_entry _ -> true
    | Instance_variable_entry ive -> ive.instance_variable_entry_is_static
    | Interface_entry _ 
    | Class_entry _
    | Package_entry _
    | Local_variable_entry _ -> false
;;

let rec get_static_env ce =
  let e = List.filter is_static_field ce.class_entry_contents in
  match ce.class_entry_extends with
    | None -> e
    | Some c -> 
	assert (ce.class_entry_name <> "Object");
	e@(get_static_env c)
;;
*)

let get_static_env ce = get_class_env ce


let rec search_name env id =
(*
  Format.eprintf "SEARCHING %s@." id;
*)
  match env with
    | [] -> 
	let le = Hashtbl.find logic_functions id in
	(le.logic_entry_return_type, 
	 mk_expr_no_loc (Logic_call(le,[])))
    | (x,Method_entry y)::r when x=id->
	begin
	  match y with
	    | [] -> assert false
	    | x::_ -> 
		begin
		  match x.method_entry_class_or_interface with
		    | Interface_entry(te) ->
			(Interface_type(te),
			 mk_expr_no_loc (Static_interface(te)))
		    | Class_entry(te) ->
			(Class_type(te),
			 mk_expr_no_loc (Static_class(te)))
		    | _ -> assert false
		end			
	end
    | (x,Logic_entry le)::r when x=id ->
	assert false
    | _::r -> search_name r id 
;;

let rec is_smallest_element me l =
  let params = me.method_entry_routine.routine_entry_parameters in
  match l with 
    | [] -> true
    | hd::tl ->
       if
         List.for_all2 
           (fun p1 p2 -> 
              match (p1, p2) with
                | ((_, Local_variable_entry lve1), (_, Local_variable_entry lve2)) -> 
                   is_subtype lve1.local_variable_entry_type lve2.local_variable_entry_type
                | _ -> assert false (* shoud never happen *)) 
           params 
           hd.method_entry_routine.routine_entry_parameters
       then is_smallest_element me tl
       else false 
;;

let find_most_specific loc l =
  match l with
    | [] -> 
	report_typing_error loc "no applicable method found";
        (* assert false *)
	raise Not_found
    | [x] -> x
    | _ ->
        (* NR - first criterion : is there a smallest me for the subtyping relation ? *)
        try 
          List.find (fun me -> is_smallest_element me l) l
        with 
          Not_found ->
	    report_typing_error loc 
	      "several applicable method found, this kind of overloading is not supported";
	     (* assert false *)
	    raise Not_found
;;

let rec search_overloaded_method loc targs l acc =
  match l with
    | [] -> find_most_specific loc acc
    | me::r ->
	let params = me.method_entry_routine.routine_entry_parameters in
	if List.length params <> List.length targs
	then search_overloaded_method loc targs r acc
	else
	  if List.for_all2
	    (fun (targ,_) param ->
	       match param with
		 | (_,Local_variable_entry lve) ->
		     is_subtype targ lve.local_variable_entry_type
		 | _ -> assert false)
	    targs params
	  then search_overloaded_method loc targs r (me::acc)
	  else search_overloaded_method loc targs r acc
	  

(* NR - JLS 15.25 *)
(* CORRECT ??? what is exactly 'a constant expression of type int whose value is representable in type byte, short or char' ??? *)
let are_numeric_types_special t1 t2 e1 e2 =
  match (t1, t2) with
  | (Byte_type, Short_type) | (Short_type, Byte_type) -> Some Short_type
  | (Short_type, Integer_type) -> 
     begin
       match e2.expr_node with
       | Lit(Int i) ->
	   if - 32768 <= i && i <= 32767 (* CORRECT ? *)
	   then Some Short_type
	   else None
       | _ -> None (* CORRECT ? *)
     end
  (* TO DO : other cases *)
  (* case t1 = t2 already handled *)
  | _ -> None
;;

let rec type_expr env e =
  match e.expr_node with
    | Lit(l) -> (type_lit l,e)
    | Bin(e1, op, e2) ->
        let (t1, te1) = type_expr env e1
        and (t2, te2) = type_expr env e2
        in
        let (t, te1', op, te2') = 
          type_bin_op 
	    e.expr_loc
            (fun t -> mk_expr_no_loc (Un("float_of_int",t)))
            op t1 t2 te1 te2
        in 
(*
  if op="||"
  then Format.fprintf Config.log "Typed: %a||%a@."
  fprintf_expr te1' fprintf_expr te2' ;	
 *)
	begin
	  match (te1.expr_node,te2.expr_node) with
	  | (Typeof te, Typed_type tt) ->
		begin
		match op with
		  | "eq_type" -> 
		      (t,{e with expr_node = Typed_typeof(te,tt)})
		  | "le_type" ->
		      (t,{e with expr_node = Typed_instanceof(te,tt)})
		  | _ -> 
		    report_typing_error e.expr_loc 
                      "operator on types is incorrect";
		    (Null_type,e)
		end
	    | (_,_) ->
		(t,{e with expr_node = Bin(te1',op,te2')})
	  end
    | Un(op,e1) ->
        let (t1,te1) = type_expr env e1 in
	let (t0,te0) =
          type_un_op 
	    e.expr_loc
	    (fun op t -> mk_expr_no_loc (Un(op,t))) op t1 te1
	in
(*
	if op="!"
	then Format.fprintf Config.log "Typed: ! %a@."
	  fprintf_expr te0;	
*)
	(t0,te0)
    | PrePostIncrDecr(op,e1) ->
        let (t1,te1) = type_expr env e1 in
	begin
	  match te1.expr_node with
	  | Typed_var(lve) ->
              if not lve.local_variable_entry_is_ref &&
		not (List.mem_assoc lve.local_variable_entry_name !current_routine.routine_entry_parameters_as_local_var)
              then 
		!current_routine.routine_entry_parameters_as_local_var <- 
		  (lve.local_variable_entry_name, Local_variable_entry(lve))::!current_routine.routine_entry_parameters_as_local_var; 
	  | _ -> ()
	end;
        type_prepost_op 
	  e.expr_loc
	  (fun op t -> mk_expr_no_loc (PrePostIncrDecr(op,t))) op t1 te1
    | Var((l,x) as v) -> 
        begin
          try
            let ve = get_name env [v] in
            match ve with
              | Local_variable_entry(lve) ->
                  (lve.local_variable_entry_type,
		   {e with expr_node = Typed_var(lve)})
              | Constant_entry(c) ->
                  (c.constant_entry_type,
                     {e with expr_node = 
		       Field_access
			 (Typed_constant_access c)})
              | Instance_variable_entry(ive) ->
                  if ive.instance_variable_entry_is_static
                  then
                    let static_ive =
                      match ive.instance_variable_entry_class_or_interface with
                       | Class_entry ce -> Static_class ce
                       | Interface_entry ie -> Static_interface ie
                       | _ -> assert false (* should never occur *)
                    in
                    (ive.instance_variable_entry_type,
                     {e with expr_node = 
		       Field_access
			 (Typed_primary_access
			    (mk_expr e.expr_loc static_ive, ive))})
                  else
                    (ive.instance_variable_entry_type,
                     {e with expr_node = 
		       Field_access
			 (Typed_primary_access
			    (snd (type_expr env (mk_expr e.expr_loc This)),ive))})
              | Interface_entry(te) ->
                  (Interface_type(te),
		   {e with expr_node = Static_interface(te)})
	      | Class_entry(te) ->
		  (Class_type(te),
		   {e with expr_node = Static_class(te)})
              | _ -> 
                  report_typing_error  l
                    (sprintf "Don't know what to do with %s in Typing.type_expr" x);
                  (Null_type,e)
          with Exit -> (Null_type,e)
        end
    | Typed_var(_) -> assert false
    | Static_class(_) -> assert false
    | Static_interface(_) -> assert false
    | This ->
        begin
          try
            let ve = List.assoc "this" env in
            match ve with
              | Local_variable_entry lve ->
                  (lve.local_variable_entry_type,
		   {e with expr_node = Typed_var(lve)})
              | _ -> assert false
          with Not_found -> 
	    report_typing_error e.expr_loc "'this' does not exist in this context";
	    (Null_type,e)
        end
    | Assign_name(x,op,e1) ->
        begin
          try
            let ve = get_name env x in
            match ve with
              | Local_variable_entry lve ->
                  let (t,te1) = type_expr env e1 in
                  if not lve.local_variable_entry_is_ref &&
		    not (List.mem_assoc lve.local_variable_entry_name !current_routine.routine_entry_parameters_as_local_var)
                  then 
		    !current_routine.routine_entry_parameters_as_local_var <- 
		      (lve.local_variable_entry_name, Local_variable_entry(lve))::!current_routine.routine_entry_parameters_as_local_var; 
		    let (op,te2) = 
		      type_assignment_op 
			e.expr_loc
			(fun t -> mk_expr_no_loc (Un("float_of_int",t)))
			op lve.local_variable_entry_type t te1
		    in			  
		    (t,
		     {e with expr_node = Assign_local_var(lve,op,te2)})
                  (* else
                    begin
                      report_typing_error  l 
			("var " ^ id ^ " is not supposed to be modifiable");
                      (Null_type,e)
                      end *)
              | Instance_variable_entry ive ->
                  let (t,te1) = type_expr env e1 in
		  let (op,te2) = 
		    type_assignment_op 
		      e.expr_loc
		      (fun t -> mk_expr_no_loc (Un("float_of_int",t)))
		      op ive.instance_variable_entry_type t te1
		  in
		    let acc = 
		      (match x with
			 | [] -> assert false
			 | [_] -> 
			     if ive.instance_variable_entry_is_static
			     then 
                               mk_expr_no_loc
                                 begin
			         match ive.instance_variable_entry_class_or_interface with
                                   | Class_entry ce -> Static_class ce
                                   | Interface_entry ie -> Static_interface ie
				   | _ -> assert false (* should never occur *)
                                 end
			     else snd(type_expr env (mk_expr_no_loc This))
			 | _::q -> snd(type_expr env (expand_name q)))
		    in
                    let temp = 
                      Assign_field
                        (Typed_primary_access(acc,ive),op,te2)
                    in 
                    (* Format.fprintf Format.err_formatter "%a@." 
                       Ast.fprintf_expr temp;
                    *)
                    (t,{e with expr_node = temp})		    
              | _ -> assert false
          with Exit ->
            (Null_type,e)
        end
    | Assign_local_var(_,_,_) -> assert false
    | Assign_field(f,op,e1) -> 
        let (t1,tf) = type_field_access env f 
        and (t2,te1) = type_expr env e1 
        in 
        let (op,te2) = 
	  type_assignment_op 
	    e.expr_loc
	    (fun t -> mk_expr_no_loc (Un("float_of_int",t)))
	    op t1 t2 te1
	in
	(t1,{e with expr_node = Assign_field(tf,op,te2)})

    | Assign_array(e1,e2,"=",e3) -> 
        let (t1,te1) = type_expr env e1
        and (t2,te2) = type_expr env e2
        and (t3,te3) = type_expr env e3
        in 
        begin
          match t1 with
            | Array_type(t) ->
                if t2 <> Byte_type && t2 <> Short_type && t2 <> Integer_type then
		  begin
		    report_typing_error  e2.expr_loc
		      "index in array assignment must be an integer";
		    (Null_type,e)
		  end
		else
                  if is_subtype t3 t || 
                     (t3 = Short_type && t = Byte_type) || 
                     (t3 = Integer_type && t = Byte_type) || 
                     (t3 = Integer_type && t = Short_type)
                  then (t,{e with expr_node = 
			   Typed_assign_array(t,te1,te2,"=",te3)})
                  else 
		    begin
		      report_typing_error  e.expr_loc
			(Format.fprintf Format.str_formatter 
			   "assignment of a value of type %a in an array of type %a" 
			   fprintf_java_type t3 
			   fprintf_java_type t1;
			Format.flush_str_formatter()) ;
		      (Null_type,e)
		  end
            | _ -> assert false
        end
    | Assign_array(e1,e2,op,e3) ->
        let (t1, te1) = type_expr env e1
        and (t2, te2) = type_expr env e2
        and (t3, te3) = type_expr env e3
        in
        begin
          match t1 with
            | Array_type(t) ->
                if t2 <> Byte_type && t2 <> Short_type && t2 <> Integer_type then
                  begin
		    report_typing_error  e2.expr_loc
		      "index in array assignment must be an integer";
		    (Null_type,e)
		  end
                else
                  let (op, te4) = 
	            type_assignment_op 
	              e.expr_loc
	              (fun t -> mk_expr_no_loc (Un("float_of_int",t)))
	              op t t3 te3
	          in (t, { e with expr_node = Typed_assign_array(t, te1, te2, op, te4)})           
            | _ -> assert false
        end
    | Typed_assign_array(_) -> assert false
    | If_expr(e1, e2, e3) -> 
        let (t1, te1) = type_expr env e1 in
	begin
          match t1 with
	  | Boolean_type | Prop_type (* NR - for 'if' expr in annotations *) ->
              let (t2, te2) = type_expr env e2
              and (t3, te3) = type_expr env e3
              in
              if t2 = t3
              then
		(t2, {e with expr_node = If_expr(te1, te2, te3)})
              else
		begin
		  (* NR - from JLS 15.25 *)
		  let ot = are_numeric_types_special t2 t3 te2 te3 in
		  match ot with
		  | Some t ->
		      (t, {e with expr_node = If_expr(te1, te2, te3)})              
			(* NR - TO DO : other cases *)
		  | None -> 
		      begin
			report_typing_error  e.expr_loc 
			  "If alternatives must be of the same type";
			(Null_type, e)           
		      end
		end
	  | _ ->
              begin
		report_typing_error  e.expr_loc 
		  "If condition must be boolean";
		(Null_type, e)
              end
	end
    | Field_access(fa) -> 
        let (t,tfa) = type_field_access env fa in
        (t,{e with expr_node = Field_access(tfa)})   
    | Typed_method_call _ -> assert false
    | Logic_call _ -> assert false
    | Method_call(ecall,(loc,id),args) -> 
	let type_to_search,te =
	  begin
	    match ecall with
	      | None ->
		  begin
		    try
		      search_name env id
		    with
			Not_found ->
			  begin
                            report_typing_error loc
                              (sprintf "No method %s found" id);
                            (Null_type,mk_expr_no_loc This)
			  end
		  end
	      | Some e ->
		  type_expr env e
	  end
	in 
        let targs = List.map (type_expr env) args in      
        begin
	  match type_to_search with
          | Class_type(ce) ->
              begin
                try
                  match List.assoc id (get_class_env ce) with
                    | Method_entry(l) -> 
                        (* TODO : overriding ?
			   check accessibility
                        *)
			let me =
			  search_overloaded_method loc targs l []
			in
			let te =
			  if not me.method_entry_is_static &&
			    ecall=None
			  then
			    snd(type_expr env (mk_expr_no_loc This))
			  else te
			in
                          (me.method_entry_return_type,
                           {e with expr_node = 
			     Typed_method_call(te,me,
					       List.map snd targs)})
                    | _ -> assert false
                with
                    Not_found ->
                      begin
                        report_typing_error  loc
                          (sprintf "No method %s found in class %s" id
                             ce.class_entry_name);
                        (Null_type,e)
                      end
              end
          | Interface_type(ce) ->
              begin
                try
                  match List.assoc id (get_interface_env ce) with
                    | Method_entry(l) -> 
                        (* TODO : overriding ?
			   check accessibility
                        *)
			let me =
			  search_overloaded_method loc targs l []
			in
			let te =
			  if not me.method_entry_is_static &&
			    ecall=None
			  then
			    snd(type_expr env (mk_expr_no_loc This))
			  else te
			in
                          (me.method_entry_return_type,
                           {e with expr_node = 
			     Typed_method_call(te,me,
					       List.map snd targs)})
                    | _ -> assert false
                with
                    Not_found ->
                      begin
                        report_typing_error  loc
                          (sprintf "No method %s found in interface %s" id
                             ce.interface_entry_name);
                        (Null_type,e)
                      end
              end
	  | Logic_type _ | Prop_type | Integer_type | Float_type | Real_type ->
	      begin
		match te.expr_node with
		  | Logic_call(le,[]) ->
		      (le.logic_entry_return_type,
		       {e with expr_node = 
			   Logic_call(le,List.map snd targs)})
		  | _ -> assert false
	      end
          | _ ->
		begin
                  report_typing_error  loc 
                    (Format.fprintf Format.str_formatter 
		       "cannot call method %s of something of type %a"
		       id fprintf_java_type type_to_search;
		    Format.flush_str_formatter());
                  (Null_type,e)
                end
        end
    | Super_method_call((loc, id), args) -> 
        begin
          try
            let type_to_search, te =
              let ve = List.assoc "this" env in
                match ve with
                 | Local_variable_entry lve -> (lve.local_variable_entry_type, mk_expr_no_loc (Typed_var(lve)))
                 | _ -> assert false (* should never happen *)
            in
            let targs = List.map (type_expr env) args in      
            begin
	      match type_to_search with
                | Class_type(ce) ->
                    begin
                      match ce.class_entry_extends with
                        | None -> 
                            begin
                              report_typing_error loc
                                (sprintf "Class %s has no superclass (it must be Object): cannot invoke 'super'" ce.class_entry_name);
                              (Null_type, e)
                            end
                        | Some ce ->
                            try
                              match List.assoc id (get_class_env ce) with
                                | Method_entry(mel) -> 
			            let me = search_overloaded_method loc targs mel [] in
                                    begin
                                      match me.method_entry_accessibility with
                                        | `PRIVATE ->
                                            begin
                                              report_typing_error loc
                                                (sprintf "private method %s of superclass %s cannot be invoked" id ce.class_entry_name);
                                              (Null_type, e)
                                            end
                                        | _ ->
                                             (me.method_entry_return_type,
                                              {e with expr_node = 
			                         Typed_method_call(te, me, List.map snd targs)})
                                    end
                                | _ -> assert false
                            with
                              Not_found ->
                                begin
                                  report_typing_error  loc
                                    (sprintf "No method %s found in class %s" id
                                       ce.class_entry_name);
                                  (Null_type, e)
                                end
                    end
                | _ -> assert false (* should never happen *)
            end             
          with Not_found -> 
	    report_typing_error e.expr_loc "'this' does not exist in this context";
	    (Null_type, e)
        end
    | Array_creation(a) -> 
	let (t,ta) = type_array_creation env a in
	(t,{e with expr_node = Array_creation(ta)})
    | Array_access(e1,e2) ->
        let (t1,te1) = type_expr env e1
        and (t2,te2) = type_expr env e2
        in
        begin
          match t1 with
            | Array_type(t) ->
                if t2 = Byte_type || t2 = Short_type || t2 = Integer_type
                then (t,{e with expr_node = Typed_array_access(t,te1,te2)})
                else
                  begin
                    report_typing_error  e.expr_loc
                      (sprintf "Array index should be an integer");
                    (Null_type,{e with expr_node = Array_access(te1,te2)})
                  end
            | _ ->
                begin
                    report_typing_error  e.expr_loc
                      (sprintf "Not an array"); 
                  (Null_type,{e with expr_node = Array_access(te1,te2)})
                end
        end
    | Typed_array_access(_) -> assert false
    | Cast(t,e1) ->
	begin
	  try
            let (ty1,te1) = type_expr env e1 
	    and tt = get_type env t
	    in
            (tt,{e1 with expr_node = Typed_cast(tt,ty1,te1)})
	  with
	      Exit -> (Null_type,e)	      
	end
    | Typed_cast _ -> assert false
    | Class_instance_creation(qi,args) ->
        begin
          let args_types,targs = List.split (List.map (type_expr env) args) 
	  in
          try
            match get_name env qi with
              | Class_entry(ce) ->
		  begin
		    try
		      let constr = 
			find_constructor ce.class_entry_constructors
			  args_types
		      in
                      (Class_type(ce),
                       {e with expr_node = 
			 Typed_class_instance_creation(constr,targs)})
		    with
			Not_found ->
			  report_typing_error e.expr_loc 
                            (sprintf "no applicable %s constructor" 
			       ce.class_entry_name);
                          (Null_type,e)
		  end
              | _ -> 
                  begin
                    match qi with
                      | [] -> assert false
                      | (loc,id)::_ ->
                          report_typing_error  loc 
                            (sprintf "%s is not a class" id);
                          (Null_type,e)
                  end
          with
	    | Not_found -> assert false
	    | Exit -> (Null_type,e)
        end
    | Typed_class_instance_creation(_,_) -> assert false
    | Instanceof(e1, t) -> 
	let (t1, te1) = type_expr env e1 in
	begin
	  match t1 with
	    | Class_type ce ->
		begin
		  try
		    (Boolean_type,
		     {e with expr_node = Typed_instanceof(te1, get_type env t)})
		  with
		      Exit -> (Null_type, e)
		end
	    | Interface_type ie ->
		begin
		  try
		    (Boolean_type,
		     {e with expr_node = Typed_instanceof(te1, get_type env t)})
		  with
		      Exit -> (Null_type, e)
		end
            | Array_type t2 ->
		begin
		  try
		    (Boolean_type,
		     {e with expr_node = Typed_instanceof(te1, get_type env t)})
		  with
		      Exit -> (Null_type,e)
		end		
	    | _ -> 
		report_typing_error  e.expr_loc 
                  "argument of instanceof is not a object";
		(Null_type,e)
	end
    | Typed_instanceof(_,_) -> assert false
    | Result -> 
        begin
          try
            let ve = List.assoc "\\result" env in
            match ve with
              | Local_variable_entry(lve) ->
                  (lve.local_variable_entry_type,
		   {e with expr_node = Typed_var(lve)})
              | _ -> assert false
          with
              Not_found -> assert false
        end
    | Old(e) -> 
        let (t,e') = type_expr env e in (t,{e with expr_node = Old(e')})
    | Fresh(e) -> 
        let (t,e') = type_expr env e in 
	(Boolean_type,{e with expr_node = Fresh(e')})
    | Not_conditionally_updated(e1, e2, e3) ->
        let (t1, te1) = type_expr env e1 in
        let (t2, te2) = type_expr env e2 in
        let (t3, te3) = type_expr env e3 in
          if (t1 = Array_type (Integer_type) || t1 = Array_type (Short_type) || t1 = Array_type (Byte_type)) && 
             (t2 = Integer_type || t2 = Array_type (Short_type) || t2 = Array_type (Byte_type)) && 
             (t3 = Integer_type || t3 = Array_type (Short_type) || t3 = Array_type (Byte_type))
          then
            (Boolean_type, { e with expr_node = Not_conditionally_updated(te1, te2, te3) })
          else
            begin
	      report_typing_error e.expr_loc
	        "arguments of \\not_conditionally_updated must be int[], int, int";
	      (Null_type, e)
	    end
    | Type(te) ->
        let tt = get_type env te in
	(Type_type,{e with expr_node = Typed_type(tt)})
    | Typed_type(_) -> assert false
    | Typeof(e1) -> 
        let (t,e') = type_expr env e1 in 
	(Type_type,{e with expr_node = Typeof(e')})
    | Typed_typeof(_,_) -> assert false
    | Quantifier(q, t, idl, e1, e2) -> 
        let ty = get_type env t in
        let (local_env, vars) = 
          List.fold_left 
            (fun (env, vars) id -> 
	      let (loc, name, ty) = get_var_id_and_type ty id in
	      let e = { 
		local_variable_entry_name = name;
                local_variable_entry_trans_name = name;
                local_variable_entry_is_ref = false;
                local_variable_entry_type = ty;
              } 
	      in
              ((name, Local_variable_entry e)::env, e::vars)) 
            (env, []) 
	    idl
        in
        let (t1, te1) = type_expr local_env e1
        and (t2, te2) = type_expr local_env e2
        in
        if is_bool_or_prop t1 && is_bool_or_prop t2 then
	  (Prop_type,
	   { e with expr_node = Typed_quantifier(q, vars, te1, te2) })
	else
	  begin
	    report_typing_error  e.expr_loc 
	      (sprintf "\\forall arguments must be boolean");
	    (Null_type,e)
          end
    | Typed_quantifier(_) -> assert false (* should never happen *)
    | Fpi(e,f1,f2) -> 
        let (t,e') = type_expr env e in 
	(Boolean_type,{e with expr_node = Fpi(e',f1,f2)})

and type_array_creation env a = 
  match a with
    | Explicit_array_creation(e,a) ->
        let (t,te) = type_expr env e in
        if t = Byte_type || t = Short_type || t = Integer_type
        then 
	  let (t,ta) = type_array_creation env a in
	  (Array_type(t),Explicit_array_creation(te,ta))
        else
          begin
            report_typing_error  e.expr_loc
              (sprintf "Array index should be an integer");
            (Null_type,a)
          end
    | Implicit_array_creation(t) -> 
	let u = get_type env t in
	(u,Typed_implicit_array_creation(u))
    | Typed_implicit_array_creation(_) -> assert false (* never happens *)

(* t is the type of an expr e and is not ArrayType,  *)
(* checks that e.id is well-typed and returns its type and the fully annotated field *)
   
and type_field_access env f =
  match f with
    | Super_access((loc,id)) -> assert false
    | Typed_primary_access(_,_) -> assert false
    | Typed_constant_access(_) -> assert false
    | Primary_access(e,(loc,id)) ->
        let (t,te) = type_expr env e in
        begin
          match t with
            | Array_type(t) -> 
                if id = "length" 
                then (Integer_type,Array_length(te))
                else 
                  begin
                    report_typing_error  loc 
                      (sprintf "%s is not an array field" id);
                    (Null_type,f)
                  end
            | Class_type ce -> 
		begin
		  try 
                    match List.assoc id (get_class_env ce) with
                      | Instance_variable_entry ive -> 
                          (ive.instance_variable_entry_type, Typed_primary_access(te, ive))
                      | Constant_entry ce ->
                          (ce.constant_entry_type, Typed_constant_access(ce))
                      | _ -> assert false (* should never happen *)
                  with
                    Not_found ->
	              report_typing_error loc 
                        ((Format.fprintf 
                            Format.str_formatter 
		            "no field %s found in class %s"
		            id ce.class_entry_name);
		         Format.flush_str_formatter());
		      (Null_type,f)
		end
	    | Interface_type(ie) ->
		begin
		  try
		    match List.assoc id (get_interface_env ie) with
		      | Constant_entry(c) -> 
			  (c.constant_entry_type,Typed_constant_access(c))
                      | Instance_variable_entry ive -> 
                          (ive.instance_variable_entry_type, Typed_primary_access(te, ive))
		      | _ -> assert false
		  with
                      Not_found -> 
			begin
			  Format.printf "interface is: %a@." fprintf_interface_entry ie;
			  report_typing_error loc 
			    ((Format.fprintf Format.str_formatter 
				"no field %s found in interface %s"
				id ie.interface_entry_name);
			     Format.flush_str_formatter());
                          (Null_type,f)
                        end
		end
	    | _ -> 
                report_typing_error loc 
                  ((Format.fprintf Format.str_formatter 
		      "cannot access field %s of something of type %a"
		      id fprintf_java_type t);
		   Format.flush_str_formatter());
		(Null_type,f)
	end
    | Array_length(_) -> assert false (* should never happen *)
;;

let rec type_set_ref env loc typv sr = 
  match sr with 
    | Set_fieldraw(f) -> 
	begin
	  let (loc,id) = f in 
	    match typv with
	      | Class_type ce ->
                  begin
		  try 
                    match List.assoc id (get_class_env ce) with
                      | Instance_variable_entry ive -> 
                          (ive.instance_variable_entry_type, Set_field(ive))
                      | _ -> assert false (* should never happen *)
                  with
                    Not_found ->
	              report_typing_error loc 
                        ((Format.fprintf 
                            Format.str_formatter 
		            "no field %s found in class %s"
		            id ce.class_entry_name);
		         Format.flush_str_formatter());
		      raise Exit
                  end
	      | _ -> 
                  report_typing_error loc 
                    ((Format.fprintf Format.str_formatter 
			"cannot access field %s of something of type %a"
			id fprintf_java_type typv);
		     Format.flush_str_formatter());
                  Null_type, Set_fieldraw(f)
	end
    | _ -> match typv with 
	  Array_type(tres) -> 
	    begin
	      match sr with 
		| Set_array_index (e) -> 
		    let (t,te) = type_expr env e 
		    in 
		      if t = Byte_type || t = Short_type || t = Integer_type then tres,Set_array_index (te)
		      else  
			begin
			report_typing_error e.expr_loc
			  "index must be integer";
			  tres,sr
			end
		| Set_array_interval(e1,e2) ->
		    let (t1,te1) = type_expr env e1
		    and (t2,te2) = type_expr env e2
		    in 
		      if t1 = Byte_type || t1 = Short_type || t1 = Integer_type then
		      if t2 = Byte_type || t2 = Short_type || t2 = Integer_type then
			tres,Set_array_interval(te1,te2)
		      else
			begin
			  report_typing_error e2.expr_loc
			    "index must be integer";
			  tres,sr
			end
		      else
			begin
			  report_typing_error e1.expr_loc
			    "index must be integer";
			  tres,sr
			end
		| Set_array -> tres,sr
		| _ -> assert false
	    end
	| _ -> 
	    begin
	      report_typing_error loc
		"assignable must have an array type ";
	      Null_type,sr
	    end
;;


let rec type_assignable_obj env o =
  match o with
    | Assignable_everything | Assignable_nothing -> o
    | Typed_assignable_set_ref _ -> assert false
    | Assignable_set_ref(e,[]) ->
	let (t,te) = type_expr env e in
	  begin
	    match te.expr_node with
	      | Field_access(Typed_primary_access(e,f)) ->
		  Typed_assignable_set_ref(t,e,[Set_field(f)])
	      | Typed_array_access(t,e1,e2) -> 
		  Typed_assignable_set_ref(Array_type(t),e1,[Set_array_index(e2)])
 	      | _ -> 
		  begin 
		    report_typing_error e.expr_loc "expr is not an assignable";o
		  end
	  end
    | Assignable_set_ref(e,l) ->
	let (t,te) = type_expr env e in
	let rec check_set_refs t lacc = function
	    [] -> List.rev lacc
	  | sr::rest -> let (t',tst) = type_set_ref env e.expr_loc t sr in
	                check_set_refs t' (tst::lacc) rest
	in Typed_assignable_set_ref(t,te,check_set_refs t [] l)
    | Assignable_all_fields(e) ->
	let (t,te) = type_expr env e in
	Assignable_all_fields(te)
;;

let type_assignable_clause env c =
  List.map
    (fun (obj, guard) ->
       let tobj = type_assignable_obj env obj in
       match guard with
	 | None -> (tobj,None)
	 | Some e ->
	     let (t,te) = type_expr env e in
	       if t = Boolean_type then
		 (tobj,Some te)
	       else
		 begin
		   report_typing_error e.expr_loc
		     "assignable guard must be boolean";
		   (tobj,Some e)
		 end)
    c
;;

let type_signal_clause env x =
  (* utile ?? *)
  if x.signals_post = expr_false then None 
  else
    (**)  
    let ty = get_type env x.signals_exception in
    let excepvar,env =
      match x.signals_variable with
	| None -> None,env
	| Some (loc,v) ->
	    let e = 
	      { local_variable_entry_name = v;
                local_variable_entry_trans_name = v;
		local_variable_entry_is_ref = false;
		local_variable_entry_type = ty;
	      }
	    in
	    (Some e,(v,Local_variable_entry e)::env)
    in
    begin
      let e = x.signals_post in
      let t,te = type_expr env e
      in
      if is_bool_or_prop t
      then 
	Some (ty,excepvar,te)
      else
	begin
	  report_typing_error  e.expr_loc
	    "'signals' assertion must be boolean";
	  None
	end
    end
;;

let type_method_specification env s purity =
  let s = Jml_desugaring.desugar s purity in
  match s with
    | Compact(compact) ->
	
	let requires = compact.compact_requires 
	and ensures = compact.compact_ensures 
	and model_ensures = compact.compact_model_ensures 
        and cardtear = 
          match compact.compact_cardtear with
          | None -> mk_expr_no_loc (Lit(Bool(true))) 
          | Some(e) -> e
 	in
	let (rtyp,tr) = type_expr env requires
	and (etyp,te) = type_expr env ensures 
	and (metyp,tme) = type_expr env model_ensures 
        and (ctyp, tc) = type_expr env cardtear
	in
	if not (is_bool_or_prop rtyp)
	then
	  begin
	    report_typing_error  requires.expr_loc
	      "'requires' assertion must be boolean";
	    s
	  end
	else
	  if not (is_bool_or_prop etyp)
	  then 
	    begin
	      report_typing_error  ensures.expr_loc
		"'ensures' assertion must be boolean";
	      s
	    end
          else 
	    if not (is_bool_or_prop ctyp)
	    then 
	      begin
	        report_typing_error  ensures.expr_loc
		  "'cardtear' assertion must be boolean";
	        s
	      end
	    else
	      let ts = type_signal_clause env compact.compact_signals
	      in
	      let ta = type_assignable_clause env compact.compact_assignable
	      in
	      Typed_compact
	        {
		  typed_compact_decls = compact.compact_decls ;
		  typed_compact_requires = tr ;
		  typed_compact_assignable = ta;
		  typed_compact_model_ensures = tme ;
		  typed_compact_ensures = te ;
		  typed_compact_signals = ts ;
                  typed_compact_cardtear = 
                    match compact.compact_cardtear with
                    | None -> None
                    | Some(_) -> Some tc; 
	        }
    | Compact_nested cns ->
        let tcns = 
	    List.map
	    (fun cs ->
	       let requires = cs.compact_requires 
	       and ensures = cs.compact_ensures 
	       and model_ensures = cs.compact_model_ensures 
              and cardtear = 
		match cs.compact_cardtear with
		| None -> mk_expr_no_loc (Lit(Bool(true))) 
		| Some(e) -> e
 	      in
	      let (rtyp,tr) = type_expr env requires
	      and (etyp,te) = type_expr env ensures 
	      and (metyp,tme) = type_expr env model_ensures 
              and (ctyp, tc) = type_expr env cardtear
	      in
	      if not (is_bool_or_prop rtyp)
	      then
		begin
		  report_typing_error  requires.expr_loc
		    "'requires' assertion must be boolean";
		  exit 1
		end
	      else
		if not (is_bool_or_prop etyp)
		then 
		  begin
		    report_typing_error  ensures.expr_loc
		      "'ensures' assertion must be boolean";
	            exit 1
		  end
		else 
		  if not (is_bool_or_prop ctyp)
		  then 
		    begin
	              report_typing_error  ensures.expr_loc
			"'cardtear' assertion must be boolean";
	              exit 1
		    end
		  else
		    let ts = type_signal_clause env cs.compact_signals
		    in
		    let ta = type_assignable_clause env cs.compact_assignable
		    in
	              { typed_compact_decls = cs.compact_decls ;
			typed_compact_requires = tr ;
			typed_compact_assignable = ta;
			typed_compact_model_ensures = tme ;
			typed_compact_ensures = te ;
			typed_compact_signals = ts ;
			typed_compact_cardtear = 
			begin 
			  match cs.compact_cardtear with
			  | None -> None
			  | Some(_) -> Some tc end})
            (cns.compact_nested_nre::cns.compact_nested_fs)
	  in
	  Typed_compact_nested
	    { typed_compact_nested_nre = List.hd tcns;
	      typed_compact_nested_fs = List.tl tcns }
    | Non_extending scl -> assert false
    | Extending_also scl -> assert false
    | Extending_and csl -> assert false
    | Typed_compact _ -> assert false
    | Typed_compact_nested _ -> assert false (* should never happen *)
;;


let type_statement_specification env a =
  let a = Jml_desugaring.desugar_statement_spec a in
  let requires = a.compact_requires 
  and ensures = a.compact_ensures 
  and model_ensures = a.compact_model_ensures 
  and cardtear =
    match a.compact_cardtear with
    | None -> mk_expr_no_loc (Lit(Bool(true)))
    | Some(e) -> e
  in
  let (rtyp,tr) = type_expr env requires
  and (etyp,te) = type_expr env ensures 
  and (metyp,tme) = type_expr env model_ensures 
  and (ctyp, tc) = type_expr env cardtear
  in
  if not (is_bool_or_prop rtyp)
  then
    begin
      report_typing_error  requires.expr_loc
	"'requires' assertion must be boolean";
      assert false
    end
  else
    if not (is_bool_or_prop etyp)
    then 
      begin
	report_typing_error  ensures.expr_loc
	  "'ensures' assertion must be boolean";
	assert false
      end
    else
      if not (is_bool_or_prop ctyp)
      then 
        begin
	  report_typing_error  ensures.expr_loc
	    "'cardtear' assertion must be boolean";
	  assert false
        end
      else
        let ts = type_signal_clause env a.compact_signals in
        let ta = type_assignable_clause env a.compact_assignable
        in
        { typed_compact_decls = a.compact_decls;
	  typed_compact_requires = tr ;
	  typed_compact_assignable = ta;
	  typed_compact_model_ensures = tme ;
	  typed_compact_ensures = te ;
	  typed_compact_signals = ts ; 
          typed_compact_cardtear = 
            match a.compact_cardtear with
            | None -> None
            | Some(_) -> Some tc; 
        }
;;

let type_jml_loop_annot env la =
  let (lit,tla) = type_expr env la.loop_invariant in
  if is_bool_or_prop lit
  then    
    let (ldt,tld) = type_expr env la.loop_decreases in
    if ldt = Byte_type || ldt = Short_type || ldt = Integer_type
    then          
      let lm = Jml_desugaring.desugar_assignable_clauses la.loop_modifies in
      { la with
	  loop_invariant = tla ; 
	  typed_loop_modifies = Option.map (type_assignable_clause env) lm ;
	  loop_decreases = tld ; }
    else
      begin
        report_typing_error la.loop_decreases.expr_loc
          "Loop variant must be integer";
        la
      end
  else
    begin
      report_typing_error la.loop_invariant.expr_loc
        "Loop invariant must be boolean";
      la
    end
;;


let rec type_variable_initializer loc name expected_type env l =
  match l with
    | Simple_initializer(e) -> 
        let (t,te) = type_expr env e in
        if is_subtype t expected_type || 
           (t = Short_type && expected_type = Byte_type) || 
           (t = Integer_type && expected_type = Byte_type) ||
           (t = Integer_type && expected_type = Short_type)
	then Simple_initializer(te)
	else
          begin
            report_typing_error loc
              (Format.fprintf Format.str_formatter 
		 "Wrong type of initializer: %a instead of %a"
		 fprintf_java_type t
		 fprintf_java_type expected_type;
	       Format.flush_str_formatter());		     
            l
          end
    | Array_initializer(l') -> 
	match expected_type with
	  | Array_type(type_of_elements) ->
	      let lt = 
		List.map (type_variable_initializer loc name type_of_elements env) l'
	      in
	      Array_initializer(lt)
	  | _ ->
	      begin
		report_typing_error  loc
		  (Format.fprintf Format.str_formatter 
		     "Wrong type of initializer: expecting an array instead of %a"
		     fprintf_java_type expected_type;
		   Format.flush_str_formatter());		     
		l
              end
;;



let type_var_decl env vd =
  begin
    try
      let var_type = get_type env vd.variable_type in
      List.map
        (fun vd ->
           let (loc,name,vdt,init) = get_var_decl var_type vd in
           let d = 
             { local_variable_entry_name = name;
               local_variable_entry_trans_name = name;
               local_variable_entry_is_ref = true;
               local_variable_entry_type = vdt; } 
           in      
           match vd.variable_initializer with
             | None -> (name, None, d)
             | Some x ->
                 let init = type_variable_initializer loc name vdt env x in
		 (name, Some init , d)) 
        vd.variable_decls
    with Exit -> []
  end

let rec type_statement env s =
  let sn = s.statement_node in
  let st =
    match sn with
    | Skip -> Skip
    | Expr(e) -> 
	let (t,te) = type_expr env e in 
	begin
	  if t <> Null_type
	  then	 
	    begin
	      match e.expr_node with
		| PrePostIncrDecr _ 
		| Assign_name _ | Assign_local_var _ 
		| Assign_field _ | Assign_array _ -> ()
		| _ -> 
		    Format.printf 
		      "Warning, %s: this expression should have type void@."
		      (Location.to_string e.expr_loc);
	    end;
	  Typed_expr(t,te)
	end
    | Typed_expr _  -> assert false
    | Var_decl(vd) -> assert false
    | Typed_var_decl(_,_,_) -> assert false
    | Throw(e) -> 
	let (t,te) = type_expr env e in
	(* TODO: verifier que t implemente Throwable *)
	Throw(te)
    | Typed_return _ -> assert false
    | Return (eopt) ->
        begin
          match eopt with
            | None -> Typed_return(Null_type,None)
            | Some e ->
                let (t,te) = type_expr env e in Typed_return(t,Some(te))
        end
    | Break(l) -> 
	if l="" then sn
	else
	  begin
	    report_typing_error  Location.dummy 
              "labelled break statements not yet supported";
            sn
	  end
    | Continue(l) -> 
         if l = "" 
         then sn
         else 
           begin
             report_typing_error Location.dummy
               "labelled continue statements not yet supported";
             sn
           end
    | Label(l,s) -> assert false
    | If_statement(e,s1,s2) -> 
        let (t,te) = type_expr env e in
        if t = Boolean_type 
        then 
          If_statement(te,type_statement env s1,type_statement env s2)
        else 
          begin
          report_typing_error  e.expr_loc
            "If condition must be boolean";
          sn
        end
    | While(loop_inv,e,s) -> 
        let (t,te) = type_expr env e 
        and tloop_inv = type_jml_loop_annot env loop_inv
        in
        if t = Boolean_type 
        then 
          While(tloop_inv,te,type_statement env s)
        else begin
          report_typing_error  e.expr_loc 
            "While loop condition must be boolean";
          sn
        end
    | Do(loop_inv,s,e) -> 
	let (t,te) = type_expr env e 
        and tloop_inv = type_jml_loop_annot env loop_inv
        in
        if t = Boolean_type 
        then 
          Do(tloop_inv,type_statement env s,te)
        else begin
          report_typing_error  e.expr_loc 
            "Do loop condition must be boolean";
          sn
        end
    | For(loop_inv,sl1,e,sl2,body) ->
        let (t,te) = type_expr env e 
        and tloop_inv = type_jml_loop_annot env loop_inv
        in
        if t = Boolean_type 
        then 
          For(tloop_inv, type_block env sl1, te, type_block env sl2,
              type_statement env body)
        else begin
          report_typing_error  e.expr_loc 
            "For loop condition must be boolean";
          sn
        end
    | For_decl(loop_inv, vd, e, sl2, body) ->
        let lvd_with_init = type_var_decl env vd in
        begin
          match lvd_with_init with
          | [(id, lvi, lve)] ->
              let loop_env = (id, Local_variable_entry lve)::env in
	      !current_routine.routine_entry_local_vars <- 
		(lve.local_variable_entry_name, Local_variable_entry lve)::!current_routine.routine_entry_local_vars;
              let (t,te) = type_expr loop_env e 
              and tloop_inv = type_jml_loop_annot loop_env loop_inv
              in
              if t = Boolean_type 
              then 
                For_typed_decl
                  (lve,lvi,tloop_inv, te, type_block loop_env sl2,
                   type_statement loop_env body)
              else begin
                report_typing_error  e.expr_loc 
                  "For loop condition must be boolean";
                sn
              end
          | _ -> assert false 
        end
    | For_typed_decl(_) -> assert false (* should never happen *)
    | Typed_try(_,_,_) ->  assert false (* should never happen *)
    | Try(b, catches, bopt) -> 
	let tb = type_block env b in
	let tcatches =
	  List.map
	    (fun (par, b) ->
	      let (id, ty) = get_parameter env par in
	      (* if 'id' clashes with Why reserved names, it is renamed in "e" *)
	      let trans_id = if Why.is_why_reserved id then "e" else id in
	      let lve = {
		local_variable_entry_name = trans_id;
		local_variable_entry_trans_name = trans_id;
		local_variable_entry_is_ref = false;
		local_variable_entry_type = ty;
	      } in
	      let tb = type_block ((id, Local_variable_entry lve)::env) b in 
	      (trans_id, ty, tb))
	    catches
	in
	let tbpot = Option.map (type_block env) bopt
	in Typed_try(tb, tcatches, tbpot)
    | Switch(e,cases) -> 
	let (t,te) = type_expr env e in
	if t <> Byte_type && t <> Short_type && t <> Integer_type then
	  begin
	    report_typing_error  e.expr_loc 
              "switch expression must be integer";
            sn 
	  end
	else
	  Switch(te,List.map (type_switch_case env) cases)
    | Block(b) -> make_block_node(type_block env b)
    | Synchronized(e,b) -> assert false
    | Jml_assert(e) ->
	let (t,te) = type_expr env e in	
	if is_bool_or_prop t
        then 
          Jml_assert(te)
	else begin
          report_typing_error  e.expr_loc 
            "assert must be boolean expressions";
          sn
        end
    | Jml_annot_statement(a) ->
        begin
        match a with 
        | Set_statement(e) -> 
	   let (t,te) = type_expr env e in
             Jml_annot_statement(Set_statement(te))
        end
    | Jml_ghost_var_decl(decls) ->
        assert false 
    | Annotated(spec,st) ->
	let spec' = type_statement_specification env spec in	
	let st' = type_statement env st in
        Typed_annotated(spec',st')
    | Typed_annotated _ -> assert false
  in mk_statement s.statement_loc st

and type_switch_case env (labels,b) =
  (List.map
     (fun c ->
       match c with
       | Case e ->
	   let (t,te) = type_expr env e in
	   if t <> Byte_type && t <> Short_type && t <> Integer_type then
	     begin
	       report_typing_error  e.expr_loc 
		 "case expression must be integer";
               c 
	     end
	   else
	     Case te
       | Default -> Default)
     labels,
   type_block env b)

and type_block env b =
  match b with
    | [] -> []
    | { statement_loc = loc ; statement_node = Var_decl(vd)}::l ->
        let lvd_with_init = type_var_decl env vd in
        let lvd = 
          List.map 
            (fun (id,_,e) -> 
	      !current_routine.routine_entry_local_vars <- (e.local_variable_entry_name, Local_variable_entry e)::!current_routine.routine_entry_local_vars;
	      (id, Local_variable_entry e)) 
            lvd_with_init 
        in
        [List.fold_left
           (fun acc (id,init,lve) -> mk_statement loc 
	       (Typed_var_decl(lve,init, acc)))
           (make_block(type_block (lvd@env) l))
	   lvd_with_init]
    | i::l -> (type_statement env i)::(type_block env l)
;;

let type_ice env ice =
  match ice with
    | Invoke_none -> ice
    | Invoke_this(el) -> 
        Invoke_this(List.map (fun e -> snd(type_expr env e)) el)
    | Invoke_super(el) -> 
        Invoke_super(List.map (fun e -> snd(type_expr env e)) el) 
;;

(*
let rec get_method_entry env mc =
  match env with
    | [] -> assert false (* TO DO: typing error to print *) 
    | (s, jee)::tl ->
       (* Format.printf "s: %s@." s;
          Format.printf "mc: %s@." (snd mc.method_call_name); *)
       if s = snd mc.method_call_name
       then
	 match jee with
           | Method_entry mel -> 
               let acc = 
                 List.filter 
                 (fun me -> 
		    let p = me.method_entry_routine.routine_entry_parameters in
                    List.length p = List.length mc.method_call_param_types &&
                    List.for_all2 
                    (fun te (_, jee) -> 
                       match jee with
                         | Local_variable_entry lve ->
                             Java_context.equals_java_type lve.local_variable_entry_type (get_type env te)
                         | _ -> assert false (* should never happen *)) 
                    mc.method_call_param_types p)
                 mel
               in
               begin
               match (List.length acc) with
                 | 0 -> get_method_entry tl mc
                 | 1 -> List.nth acc 0
                 | _ -> assert false (* should never happen *)
               end
           | _ -> assert false (* TO DO: typing error to print *) 
       else 
         get_method_entry tl mc
;;

let rec type_call_sequence env cs =
  match cs with
    | Simple(mc) -> Typed_simple(get_method_entry env mc)
    | Typed_simple(_) -> assert false (* should never happen *)
    | Optional(cs) -> Optional(type_call_sequence env cs)
    | Multiple(cs) -> Multiple(type_call_sequence env cs)
    | At_least_one(cs) -> At_least_one(type_call_sequence env cs)
    | Compose(cs1, cs2) -> Compose(type_call_sequence env cs1, type_call_sequence env cs2)
    | Or(cs1, cs2) -> Or(type_call_sequence env cs1, type_call_sequence env cs2)
;;
*)

let type_jml_specification env s =
  match s with
  | Invariant(i) ->
      let (t,ti) = type_expr env i in
      if is_bool_or_prop t then 
	Invariant(ti)
      else
	begin
	  report_typing_error  i.expr_loc 
            "An invariant must be a Boolean expression";
	  s
	end
  | Static_invariant(i) ->
      let (t, ti) = type_expr env i in
      if is_bool_or_prop t then 
	Static_invariant(ti)
      else
	begin
	  report_typing_error  i.expr_loc 
            "An static invariant must be a Boolean expression";
	  s
	end
  | Representation_invariant(i) ->
      let (t, ti) = type_expr env i in
      if is_bool_or_prop t then 
	Representation_invariant(ti)
      else
	begin
	  report_typing_error  i.expr_loc 
            "A representation invariant must be a Boolean expression";
	  s
	end
  | Jml_variable_declaration _ -> assert false
  | Represents((loc,id),e) -> 
      begin
	try
	  match List.assoc id env with
	  | Instance_variable_entry ive ->
	      let (t,te) = type_expr env e in
	      if t <> ive.instance_variable_entry_type then
		begin
		  report_typing_error  loc  
		    "not the same type as what it represents";
		  s
		end
	      else
		Typed_represents(ive,te)
	  | _ -> assert false
	with
	  Not_found ->
	    begin
	      report_typing_error  loc  
		"unknown identifier";
	      s
	    end
      end
  | Typed_represents _ -> assert false
(*  | Call_sequence cs -> Call_sequence(type_call_sequence env cs) *)
;;

let type_class_field class_entry env c class_name =
  match c with 
  | Typed_method(me,jml,body) ->
      let this_env =
	if me.method_entry_is_static
	then (get_static_env class_entry)@env
	else
	  let this =
	    { 
	      local_variable_entry_name = "this";
	      local_variable_entry_trans_name = "this";
	      local_variable_entry_is_ref = 
		if !Config.use_value_model then 
		  true 
		    (* TODO: false unless declared modifiable *) 
		else false;
	      local_variable_entry_type = Class_type class_entry;
	    }
	  in
          ("this",Local_variable_entry this)::
	  ((get_class_env class_entry)@env)
      in
      let local_env = 
	me.method_entry_routine.routine_entry_parameters@this_env 
      in
      let jml_env = 
        ("\\result",
         Local_variable_entry
           { local_variable_entry_type = me.method_entry_return_type; 
             local_variable_entry_name = "result";
             local_variable_entry_trans_name = "result";
             local_variable_entry_is_ref = false})
         ::local_env
      in
      if !Config.verbose >= 2 then
	Format.printf "Typing specification of %s@." me.method_entry_name;
      begin
	try
	  let tjml = type_method_specification jml_env jml me.method_entry_is_pure
	  in
	  match body with
	    | None -> Typed_method(me, tjml, None)
	    | Some b ->		
		if Config.mem_methods 
		  class_entry.class_entry_name me.method_entry_name 
		then 
		  begin
		    me.method_entry_routine.routine_entry_produce_body <- true;
		    current_routine := me.method_entry_routine;
		    Typed_method(me, tjml, Some(type_block local_env b))
		  end
		else Typed_method(me, tjml, body)
	with
	  Jml_desugaring.Desugaring_failure msg ->
	    report_typing_error (fst class_name)
              ("Warning: JML Desugaring has failed (" ^ msg ^ ")");
	    c
      end
  | Typed_constructor(ce,jml,ice,block) ->
      let this_env =
	let this =
	  { 
	    local_variable_entry_name = "this";
	    local_variable_entry_trans_name = "this";
	    local_variable_entry_is_ref = 
	      if !Config.use_value_model then 
		true 
		  (* TODO: false unless declared modifiable *) 
	      else false;
	    local_variable_entry_type = Class_type class_entry;
	  }
	in
        ("this",Local_variable_entry this)::
	((get_class_env class_entry)@env)
      in
      let local_env = 
	ce.constructor_entry_routine.routine_entry_parameters@this_env 
      in
      begin
	try
	  let tjml = type_method_specification local_env jml false in
	  let tbody = 
	    if Config.mem_methods class_entry.class_entry_name
	      class_entry.class_entry_name
	    then 
	      begin
		ce.constructor_entry_routine.routine_entry_produce_body <- true;
		current_routine := ce.constructor_entry_routine;
		type_block local_env block 
	      end
	    else block
	  in	      
	  Typed_constructor(ce,tjml,type_ice local_env ice,tbody)
	with
	  Jml_desugaring.Desugaring_failure msg ->
	    report_typing_error (fst class_name)
              ("Warning: JML Desugaring has failed (" ^ msg ^ ")");
	    c
      end
  | Constructor(c_decl) -> c
  | Typed_variable(ve, init) -> 
      begin
        match init with
          | None -> c
          | Some i ->
	      let this_env =
		let this =
		  { 
		    local_variable_entry_name = "this";
		    local_variable_entry_trans_name = "this";
		    local_variable_entry_is_ref = 
		      if !Config.use_value_model then 
			true 
			  (* TODO: false unless declared modifiable *) 
		      else false;
		    local_variable_entry_type = Class_type class_entry;
		  }
		in
		("this",Local_variable_entry this)::
		((get_class_env class_entry)@env)
	      in
              let rec get_loc i =
                match i with
                  | Simple_initializer(e) -> e.expr_loc
                  | Array_initializer(vil) ->
                      begin 
                      match vil with
                        | [] -> Location.dummy 
			| hd::tl -> get_loc hd 
                      end
              in
              let ti = 
		type_variable_initializer 
		  (get_loc i) ve.instance_variable_entry_name 
		  ve.instance_variable_entry_type this_env i 
	      in
              Typed_variable(ve,Some(ti))
      end
  | Static_initializer(block) ->
      assert false
  | Specification(s) ->
      let this_env =
	let this =
	  { 
	    local_variable_entry_name = "this";
	    local_variable_entry_trans_name = "this";
	    local_variable_entry_is_ref = 
	      if !Config.use_value_model then 
		true 
		  (* TODO: false unless declared modifiable *) 
	      else false;
	    local_variable_entry_type = Class_type class_entry;
	  }
	in
        ("this",Local_variable_entry this)::
	((get_class_env class_entry)@env)
      in
      Specification(List.map (type_jml_specification this_env) s)
  | Method (m, _) -> c
  | Variable (_) -> c
;;


let type_interface_member interface_entry env c interface_name =
  match c with 
  | Constant_declaration(var_decl) -> c
  | Method_header(m_decl) -> c
  | Typed_constant_declaration(ive, var_init) ->
      let local_env = (get_interface_env interface_entry)@env in
      begin
      match var_init with 
        | None -> Typed_constant_declaration(ive, None)
        | Some i -> 
            let rec get_loc i =
              match i with
                | Simple_initializer(e) -> e.expr_loc
                | Array_initializer(vil) -> get_loc (List.hd vil) 
            in
            let ti = 
	      Option.map
	        (type_variable_initializer 
	           (get_loc i) ive.constant_entry_name 
	           ive.constant_entry_type local_env) var_init 
            in
            Typed_constant_declaration(ive,ti)
      end 
  | Typed_method_header(me,jml_spec) -> 
      let this_env =
	if me.method_entry_is_static
	then (get_interface_env interface_entry)@env
	else
	  let this =
	    { 
	      local_variable_entry_name = "this";
	      local_variable_entry_trans_name = "this";
	      local_variable_entry_is_ref = 
		if !Config.use_value_model then 
		  true 
		    (* TODO: false unless declared modifiable *) 
		else false;
	      local_variable_entry_type = Interface_type interface_entry;
	    }
	  in
          ("this",Local_variable_entry this)::
	  ((get_interface_env interface_entry)@env)
      in
      let local_env = me.method_entry_routine.routine_entry_parameters@this_env in
      let jml_env = 
        ("\\result",
         Local_variable_entry
           { local_variable_entry_type = me.method_entry_return_type; 
             local_variable_entry_name = "result" ;
             local_variable_entry_trans_name = "result";
             local_variable_entry_is_ref = false})
         ::local_env
      in
      if !Config.verbose >= 2 then
	Format.printf "Typing specification of %s@." me.method_entry_name;
      begin
	try
	  let tjml = type_method_specification jml_env jml_spec me.method_entry_is_pure 
	  in
	  Typed_method_header(me, tjml)
	with
	  Jml_desugaring.Desugaring_failure msg ->
	    report_typing_error (fst interface_name)
              ("Warning: JML Desugaring has failed (" ^ msg ^ ")");
	    c
      end
  | Model_variable _
  | Typed_model_variable _ -> c
  | Interface_specification(s) ->
      let this_env =
	let this =
	  { 
	    local_variable_entry_name = "this";
	    local_variable_entry_trans_name = "this";
	    local_variable_entry_is_ref = 
	      if !Config.use_value_model then 
		true 
		  (* TODO: false unless declared modifiable *) 
	      else false;
	    local_variable_entry_type = Interface_type interface_entry;
	  }
	in
        ("this",Local_variable_entry this)::
	((get_interface_env interface_entry)@env)
      in
      Interface_specification(List.map (type_jml_specification this_env) s)
;;

let phase3 global_env filename comp_unit =
  let env_after_imports = 
    build_env_after_imports true global_env comp_unit  
  in
  (*
    Format.printf "****** java context after imports in file %s *******@." filename;
    printf_java_context env_after_imports;
    Format.printf "@.****** end java context *******@.";
  *)
  let new_type_decls = 
    List.map
      (fun (doc_comment,typedecl) ->
         let new_typedecl =
           match typedecl with
             | Class _ -> assert false
	     | Axiom(id,e) ->
		 let (t,te) = type_expr env_after_imports e in
		 if is_bool_or_prop t
		 then 
		   Axiom(id,te)
		 else
		   begin
		     report_typing_error e.expr_loc 
		       "An axiom must be a Boolean expression";
		     typedecl
		   end
	     | Logic_typedecl _ -> typedecl
	     | Logic_reads _ -> assert false
	     | Logic_def _ -> assert false
	     | Typed_logic_reads(le,r) -> 
		 let local_env = le.logic_entry_parameters@env_after_imports in
		 let eff = type_assignable_clause local_env r in
		 Typed_logic_reads(le,eff) 
	     | Typed_logic_def(le,e) -> 
		 let local_env = le.logic_entry_parameters@env_after_imports in
		 let (t,te) = type_expr local_env e in
(*
		 if is_bool_or_prop t
		 then 
*)
		   Typed_logic_def(le,te)
(*
		 else
		   begin
		     report_typing_error e.expr_loc 
		       "A predicate must be a Boolean expression";
		     typedecl
		   end
*)
	     | Typed_class(class_entry,class_decl) ->
		 let env = env_after_imports in
                 let new_class_fields =
                   List.map
                     (fun (doc,c) -> 
                        let tf = 
			  type_class_field class_entry env c class_decl.class_name
			in (doc,tf))
                     class_decl.class_fields
                 in
                 Typed_class
		   (class_entry,
		    { class_decl with class_fields = new_class_fields })
	     | Interface _ -> assert false
             | Typed_interface(interface_entry,interface_decl) -> 
                 let new_interface_members =
                   List.map
                     (fun (doc,c) -> 
                        let tf = type_interface_member interface_entry 
				   env_after_imports c interface_decl.interface_name
			in
                        (doc,tf))
                     interface_decl.interface_members
                 in
                 Typed_interface
		   (interface_entry,
		    { interface_decl with interface_members = 
		      new_interface_members })
         in
         (doc_comment,new_typedecl))
      comp_unit.cu_type_decls
  in
  { comp_unit with cu_type_decls = new_type_decls }
;;





